###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
"""
This module describes a theme structure for VisTrails GUI. It
specifies colors, background images and other measurements
"""
from __future__ import division

from PyQt4 import QtCore, QtGui

from vistrails.core.utils.color import ColorByName
from vistrails.core.theme import DefaultCoreTheme
import vistrails.core.system

import unittest
import vistrails.gui.qt

################################################################################

def _create_configure_shape(w, h):
    poly = QtGui.QPolygon(3)
    poly.setPoint(0, 0, 0)
    poly.setPoint(1, 0, h)
    poly.setPoint(2, w, h/2)
    return QtGui.QPolygonF(poly)
    

class DefaultTheme(DefaultCoreTheme):
    """
    This is the default theme which contains color, images,
    measurements, etc. for Vistrail. Other themes should derive from
    this class and change appropriate values
    
    """
    
    def __init__(self):
        """ DefaultTheme() -> DefaultTheme
        This is for initializing all Qt objects
        
        """
        DefaultCoreTheme.__init__(self)
        ######################
        #### MEASUREMENTS ####

        # pipeline view bounding rect
        self.BOUNDING_RECT_MINIMUM = 512

        # Port shape
        self.PORT_RECT = QtCore.QRectF(0, 0, self.PORT_WIDTH, self.PORT_HEIGHT)

        # Configure button shape
        self.CONFIGURE_SHAPE = _create_configure_shape(self.CONFIGURE_WIDTH,
                                                       self.CONFIGURE_HEIGHT)

        #### BRUSH & PEN ####
        # Background brush of the pipeline view
        # self.PIPELINE_VIEW_BACKGROUND_BRUSH = QtGui.QBrush(
        #     QtGui.QImage(vistrails.core.system.vistrails_root_directory() +
        #                  '/gui/resources/images/pipeline_bg.png'))
        #     #QtGui.QColor("white"))
        # # Background brush of the version tree
        # self.VERSION_TREE_BACKGROUND_BRUSH = QtGui.QBrush(
        #     QtGui.QImage(vistrails.core.system.vistrails_root_directory() +
        #                  '/gui/resources/images/version_bg.png'))
        # Background brush of the query pipeline view
        # self.QUERY_BACKGROUND_BRUSH = QtGui.QBrush(
        #     QtGui.QImage(vistrails.core.system.vistrails_root_directory() +
        #                  '/gui/resources/images/query_bg.png'))
        self.PIPELINE_VIEW_BACKGROUND_BRUSH = QtGui.QBrush(
            QtGui.QColor(128, 128, 128))
        self.VERSION_TREE_BACKGROUND_BRUSH = QtGui.QBrush(
            QtGui.QColor(240, 240, 240))
        self.QUERY_BACKGROUND_BRUSH = QtGui.QBrush(
            QtGui.QColor(119, 143, 159))
        self.QUERY_RESULT_BACKGROUND_BRUSH = QtGui.QBrush(
            QtGui.QColor(208, 226, 239))

        # Pen to draw a module shape at regular state
        self.MODULE_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('black')))), 2)
        # Pen to draw a module shape when selected
        self.MODULE_SELECTED_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('goldenrod_medium')))), 3)
        # Brush and pen to draw a module label
        self.MODULE_LABEL_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('black')))), 2)
        self.MODULE_LABEL_SELECTED_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('black')))), 2)
        # Brush to draw a module shape at different states
        self.MODULE_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('light_grey'))))
        self.ERROR_MODULE_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('salmon'))))
        self.SUCCESS_MODULE_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('mint'))))
        self.ACTIVE_MODULE_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('navajo_white'))))
        self.COMPUTING_MODULE_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('yellow'))))
        self.NOT_EXECUTED_MODULE_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('light_goldenrod'))))
        self.PERSISTENT_MODULE_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('slate_blue'))))
        self.SUSPENDED_MODULE_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('aureoline_yellow'))))

        self.INVALID_MODULE_PEN = QtGui.QPen(QtGui.QBrush(
                QtGui.QColor(51, 51, 51, 255)), 2)
        self.INVALID_MODULE_LABEL_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(51, 51, 51, 255)), 2)
        self.INVALID_MODULE_BRUSH = QtGui.QBrush(
            QtGui.QColor(125, 92, 92, 255))

        # Pen and brush for un-matched queried modules
        self.GHOSTED_MODULE_PEN = QtGui.QPen(QtGui.QBrush(
                QtGui.QColor(*(ColorByName.get_int('dark_dim_grey')))), 2)
        # Pen to draw module label when it is unmatched due to a query
        self.GHOSTED_MODULE_LABEL_PEN = QtGui.QPen(QtGui.QBrush(
                QtGui.QColor(*(ColorByName.get_int('dark_dim_grey')))), 2)
        self.GHOSTED_MODULE_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('light_dim_grey'))))

        # Pen and brush for breakpoint modules
        self.BREAKPOINT_MODULE_PEN = QtGui.QPen(QtGui.QBrush(
                QtGui.QColor(*(ColorByName.get_int('dark_dim_grey')))), 2)
        self.BREAKPOINT_MODULE_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('light_dim_grey'))))
        self.BREAKPOINT_MODULE_LABEL_PEN = QtGui.QPen(QtGui.QBrush(
                QtGui.QColor(*(ColorByName.get_int('dark_dim_grey')))), 2)

        # Module pen styles
        self.ABSTRACTION_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('black')))), 2,
                                          QtCore.Qt.DotLine)
        self.GROUP_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('black')))), 2,
                                          QtCore.Qt.DashLine)

        # Brush and pen to draw a port shape at regular state
        self.PORT_PEN_WIDTH_MANDATORY = 'mandatory'
        self.PORT_PEN_WIDTH_NORMAL = 'normal'
        self.PORT_PEN_WIDTH_SELECTED = 'selected'
        self.PORT_PEN_WIDTHS = {self.PORT_PEN_WIDTH_MANDATORY: 1.5, 
                                self.PORT_PEN_WIDTH_NORMAL: 1.0, 
                                self.PORT_PEN_WIDTH_SELECTED: 3.0}
        self.PORT_PEN_COLOR_NORMAL = 'normal'
        self.PORT_PEN_COLOR_FULL = 'full'
        self.PORT_PEN_COLOR_INVALID = 'invalid'
        self.PORT_PEN_COLOR_GHOSTED = 'ghosted'
        self.PORT_PEN_COLOR_SELECTED = 'selected'
        self.PORT_PEN_COLORS = \
            {self.PORT_PEN_COLOR_NORMAL: \
                 QtGui.QColor(*(ColorByName.get_int('black'))),
             self.PORT_PEN_COLOR_FULL: \
                 QtGui.QColor(*(ColorByName.get_int('dark_dim_grey'))),
             self.PORT_PEN_COLOR_INVALID: QtGui.QColor(51, 51, 51, 255),
             self.PORT_PEN_COLOR_GHOSTED: \
                 QtGui.QColor(*(ColorByName.get_int('dark_dim_grey'))),
             self.PORT_PEN_COLOR_SELECTED: \
                 QtGui.QColor(*(ColorByName.get_int('goldenrod_medium')))
             }
            
        self.PORT_PENS = {}
        for (color_type, color) in self.PORT_PEN_COLORS.iteritems():
            for (width_type, width) in self.PORT_PEN_WIDTHS.iteritems():
                self.PORT_PENS[(color_type, width_type)] = \
                    QtGui.QPen(QtGui.QBrush(color), width)

        self.PORT_CONNECTED_BRUSH = QtGui.QBrush(
            QtGui.QColor(0,0,0,60))
        self.PORT_BRUSH = QtGui.QBrush(QtCore.Qt.NoBrush)
        self.PORT_MANDATORY_BRUSH = QtGui.QBrush(QtGui.QColor(255,255,255,180))
        self.GHOSTED_PORT_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('light_dim_grey'))))
        self.INVALID_PORT_BRUSH = QtGui.QBrush(
            QtGui.QColor(125, 92, 92, 255))

        # Pen and brush for drawing the configure button
        self.CONFIGURE_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('black')))), 1)
        self.CONFIGURE_BRUSH= QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('black'))))

        # Pen and brush for drawing the ghosted configure button
        self.GHOSTED_CONFIGURE_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('dark_dim_grey')))), 2)
        self.GHOSTED_CONFIGURE_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('dark_dim_grey'))))

        # Brush and pen to draw connections
        self.CONNECTION_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('black')))), 2)
        self.CONNECTION_SELECTED_PEN = QtGui.QPen(
                QtGui.QBrush(
                    QtGui.QColor(*(ColorByName.get_int('goldenrod_medium')))),
                3,
                QtCore.Qt.SolidLine)
        self.CONNECTION_SELECTED_CONVERTING_PEN = QtGui.QPen(
                QtGui.QBrush(
                    QtGui.QColor(*(ColorByName.get_int('goldenrod_medium')))),
                3,
                QtCore.Qt.DotLine)
        self.CONNECTION_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('black'))))

        # Pen for drawing while connecting any ghosted modules
        self.GHOSTED_CONNECTION_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('dark_dim_grey')))), 2)

        # Pen to draw version tree node
        self.VERSION_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('black')))), 2)    
        self.GHOSTED_VERSION_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('light_grey')))), 2)    
        self.VERSION_SELECTED_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('goldenrod_medium')))), 4)

        self.VERSION_LABEL_COLOR = \
            QtGui.QColor(*(ColorByName.get_int('black')))
        self.GHOSTED_VERSION_LABEL_COLOR = \
            QtGui.QColor(*(ColorByName.get_int('light_grey')))

        # Brush to draw version belongs to the current user
        self.VERSION_USER_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('desatcornflower'))))
        self.GHOSTED_VERSION_USER_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('very_light_grey'))))

        # Brush to draw version belongs to the other users
        self.VERSION_OTHER_BRUSH = QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('melon'))))
    
        # Brush and pen to draw a link between two versions
        self.LINK_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('black')))), 1.5)
        self.LINK_SELECTED_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('goldenrod_medium')))), 3)
        self.GHOSTED_LINK_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*(ColorByName.get_int('light_grey')))), 2)

        # Selection box color
        self.SELECTION_BOX_BRUSH = QtGui.QBrush(
            QtGui.QColor(*ColorByName.get_int('light_grey')))
        self.SELECTION_BOX_PEN = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor(*ColorByName.get_int('lamp_black'))), 1)
        
        # Color of the version is being diff from in
        self.VISUAL_DIFF_FROM_VERSION_BRUSH = QtGui.QBrush(
            QtGui.QColor(*ColorByName.get_int('melon')))
        
        # Color of the version is being diff to in
        self.VISUAL_DIFF_TO_VERSION_BRUSH = QtGui.QBrush(
            QtGui.QColor(*ColorByName.get_int('steel_blue_light')))
        
        # Color of the paramter changed modules in Visual Diff
        self.VISUAL_DIFF_PARAMETER_CHANGED_BRUSH = QtGui.QBrush(
            QtGui.QColor(*ColorByName.get_int('light_grey')))
        
        # Color of the shared modules in Visual Diff
        self.VISUAL_DIFF_SHARED_BRUSH = QtGui.QBrush(
            QtGui.QColor(155, 155, 155, 255))

        # Color of shared modules in Visual Diff matched by heuristic
        self.VISUAL_DIFF_MATCH_BRUSH = QtGui.QBrush(
            QtGui.QColor(*ColorByName.get_int('white')))
        
        # Color of partially shared modules in Visual Diff
        self.VISUAL_DIFF_SUMMARY_BRUSH = QtGui.QBrush(
            QtGui.QColor(*ColorByName.get_int('spring_green')))

        # Pen & Brush of the circled id on the right corner of the
        # virtual cell label
        self.ANNOTATED_ID_PEN = QtGui.QPen(
            QtCore.Qt.white)
        self.ANNOTATED_ID_BRUSH = QtGui.QBrush(
            QtGui.QColor(157, 0, 0, 255))
    
        #### FONTS ####        
        # Font for module text
        # Use fixed dpi to get same font size on all platforms
        def fixDPI(i):
            return i*72//QtGui.QApplication.desktop().logicalDpiY()
        GRAPHICS_FONT = "Arial"
        self.MODULE_FONT = QtGui.QFont(GRAPHICS_FONT, fixDPI(14), QtGui.QFont.Bold)
        self.MODULE_FONT_METRIC = QtGui.QFontMetrics(self.MODULE_FONT)
        self.MODULE_DESC_FONT = QtGui.QFont(GRAPHICS_FONT, fixDPI(12))
        self.MODULE_DESC_FONT_METRIC = QtGui.QFontMetrics(self.MODULE_DESC_FONT)
        self.MODULE_EDIT_FONT = QtGui.QFont(GRAPHICS_FONT, fixDPI(10))
        self.MODULE_EDIT_FONT_METRIC = QtGui.QFontMetrics(self.MODULE_EDIT_FONT)
    
        # Font for version text
        self.VERSION_FONT = QtGui.QFont(GRAPHICS_FONT, fixDPI(15), QtGui.QFont.Bold)
        self.VERSION_FONT_METRIC = QtGui.QFontMetrics(self.VERSION_FONT)
        self.VERSION_DESCRIPTION_FONT = QtGui.QFont(GRAPHICS_FONT, fixDPI(15),
                                                    QtGui.QFont.Normal, True)
        self.VERSION_DESCRIPTION_FONT_METRIC = \
            QtGui.QFontMetrics(self.VERSION_DESCRIPTION_FONT)

        self.VERSION_PROPERTIES_FONT = QtGui.QFont("Arial", 12)
        self.VERSION_PROPERTIES_FONT_METRIC = \
            QtGui.QFontMetrics(self.VERSION_PROPERTIES_FONT)
        self.VERSION_PROPERTIES_PEN =  QtGui.QBrush(
            QtGui.QColor(20, 100, 20, 255))
            

        # Font showing on the Parameter Inspector window of Visual Diff
        self.VISUAL_DIFF_PARAMETER_FONT = QtGui.QFont('Arial', 10)
        
        # Font showing on the Legend window of Visual Diff
        self.VISUAL_DIFF_LEGEND_FONT = QtGui.QFont('Arial', 9)

        # Font for PythonSource
        self.PYTHON_SOURCE_EDITOR_FONT = QtGui.QFont('Courier', 10, 
                                                     QtGui.QFont.Normal)
        # Font for Splash Screen messages
        self.SPLASH_SCREEN_FONT = QtGui.QFont('Arial', 10,
                                              QtGui.QFont.Normal)
        #### ICONS & IMAGES ####
        #The application disclaimer image
        self.DISCLAIMER_IMAGE = QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/disclaimer.png')
        #The application icon
        self.APPLICATION_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/vistrails_icon_small.png')

        #The application pixmap
        self.APPLICATION_PIXMAP = QtGui.QPixmap(
             vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/vistrails_icon_small.png')

        # The execute icons in the first spot of vistrail view toolbar
        self.EXECUTE_PIPELINE_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/execute.png')
        self.EXECUTE_EXPLORE_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/execute_explore.png')

        # The undo icons for the vistrail view toolbar
        self.UNDO_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/undo.png')

        # The redo icons for the vistrail view toolbar
        self.REDO_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/redo.png')

        # Icon to select the tabbed view
        self.TABBED_VIEW_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/tabbed_view.png')
    
        # Icon to select the horizontal split view
        self.HORIZONTAL_VIEW_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/hor_view.png')
    
        # Icon to select the vertical split view
        self.VERTICAL_VIEW_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/ver_view.png')
    
        # Icon to select the docking-style view
        self.DOCK_VIEW_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/dock_view.png')
    
        # Toolbar icon for creating a new Vistrail
        self.NEW_VISTRAIL_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/new_vistrail.png')
        
        # Toolbar icon for opening a vistrail
        self.OPEN_VISTRAIL_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/open_vistrail.png')

        #Toolbar icon for opening a vistrail from a database
        self.OPEN_VISTRAIL_DB_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/openfromdb.png')

        #Icon for database connections
        self.DB_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/database.png')

        #Icon for vistrails files
        self.FILE_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/file.png')
        
        # Toolbar icon for save the current Vistrail
        self.SAVE_VISTRAIL_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/save_vistrail.png')
    
        # Toolbar icon for toggling console mode window
        self.CONSOLE_MODE_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/console.png')

        # Background image of the Visual Diff pipeline view
        self.VISUAL_DIFF_BACKGROUND_IMAGE = QtGui.QImage(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/pipeline_bg.png')
        
        # Toolbar icon for showing the Parameter Inspector window
        self.VISUAL_DIFF_SHOW_PARAM_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/show_params.png')
        
        # Toolbar icon for showing the Legend window
        self.VISUAL_DIFF_SHOW_LEGEND_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/show_legends.png')

        # Toolbar icon for creating an analogy
        self.VISUAL_DIFF_CREATE_ANALOGY_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/analogy.png')

        # Toolbar icon for close button on the vistrail tabbar
        self.VIEW_MANAGER_CLOSE_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/closeview.png')

        # Toolbar icon for the dock toolbutton on the splitted window
        self.DOCK_BACK_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/dockback.png')

        # Icon for adding string in the parameter exploration widget
        self.ADD_STRING_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/edit_add.png')

        # Icon for moving string up in the parameter exploration widget
        self.UP_STRING_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/up.png')

        # Icon for moving string up in the parameter exploration widget
        self.DOWN_STRING_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/down.png')

        # Icon for expand all/collapse all buttons in the Module Palette
        self.EXPAND_ALL_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/expand_all.png')
        
        self.COLLAPSE_ALL_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/collapse_all.png')
        
        # Icons for tree/list view buttons in the Workspace
        self.LIST_VIEW_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/list_view.png')
        
        self.TREE_VIEW_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/tree_view.png')

        # Toolbar icons for views
        self.PIPELINE_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/pipeline.png')
        self.HISTORY_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/history.png')
        self.QUERY_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/full_tree.png')
        self.EXPLORE_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/explore.png')
        self.PROVENANCE_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/search_database.png')

        # Toolbar icon for visual query on a vistrail
        self.VISUAL_QUERY_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/visual_query.png')
        
        # Toolbar icon for viewing the whole version tree
        self.VIEW_FULL_TREE_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/full_tree.png')

        # Toolbar icon for viewing the whole version tree
        self.PERFORM_PARAMETER_EXPLORATION_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/perform_pe.png')

        # Toolbar icon for dragging pixmap of VisDiff
        self.VERSION_DRAG_PIXMAP = QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/dragging.png')

        #Tabbar icon for detach a tab        
        self.DETACH_TAB_ICON = QtGui.QIcon()
        self.DETACH_TAB_ICON.addFile(vistrails.core.system.vistrails_root_directory() +
                                     '/gui/resources/images/detach.png',
                                     mode=QtGui.QIcon.Normal)
        self.DETACH_TAB_ICON.addFile(vistrails.core.system.vistrails_root_directory() +
                                     '/gui/resources/images/detach_on.png',
                                     mode=QtGui.QIcon.Active)

        #toolbutton icon for pin/unpin palette
        self.PINNED_PALETTE_ICON = QtGui.QIcon(
                                vistrails.core.system.vistrails_root_directory() +
                                '/gui/resources/images/pinned.png')
        self.UNPINNED_PALETTE_ICON = QtGui.QIcon(
                                vistrails.core.system.vistrails_root_directory() +
                                '/gui/resources/images/unpinned.png')
        # Parameter Exploration Pixmaps
        self.EXPLORE_COLUMN_PIXMAP = QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/column.png')
        self.EXPLORE_ROW_PIXMAP = QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/row.png')
        self.EXPLORE_SHEET_PIXMAP = QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/sheet.png')
        self.EXPLORE_TIME_PIXMAP = QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/time.png')        
        self.EXPLORE_SKIP_PIXMAP = QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/skip.png')        
        self.REMOVE_PARAM_PIXMAP = QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/remove_param.png')        
        self.RIGHT_ARROW_PIXMAP = QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/right.png')        
        self.LEFT_ARROW_PIXMAP = QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/left.png')
        
        # Cursor for zoom in/out graphics views
        self.SELECT_CURSOR = QtGui.QCursor(QtCore.Qt.ArrowCursor)
        self.OPEN_HAND_CURSOR = QtGui.QCursor(QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/open_hand.png'))
        self.CLOSE_HAND_CURSOR = QtGui.QCursor(QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/close_hand.png'))
        self.ZOOM_CURSOR = QtGui.QCursor(QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/zoom.png'))
                
        # Cursor icon for zoom in/out graphics views
        self.SELECT_ICON = QtGui.QIcon(QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/select_icon.png'))
        self.PAN_ICON = QtGui.QIcon(QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/pan_icon.png'))
        self.ZOOM_ICON = QtGui.QIcon(QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/zoom_icon.png'))
        
        # Mashup Icons
        self.EXECUTE_MASHUP_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/mashup_execute.png')
        self.MASHUP_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/mashup_create.png')
        self.MASHUP_ALIAS_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/alias.png')

        # Job View Icons
        self.JOB_SCHEDULED = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/appointment-new.png')
        self.JOB_FINISHED = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/emblem-important.png')
        self.JOB_CHECKING = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/view-refresh.png')
        
        # Saved Queries icons
        self.QUERY_VIEW_ICON = self.ZOOM_ICON
        self.QUERY_ARROW_ICON = QtGui.QIcon(QtGui.QPixmap(
                vistrails.core.system.vistrails_root_directory() +
                '/gui/resources/images/zoom_arrow_icon.png'))
        self.QUERY_EDIT_ICON = QtGui.QIcon(QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/edit.png'))

        # Icon on the button to switch to single/multi line string edition
        self.MULTILINE_STRING_ICON = QtGui.QIcon(QtGui.QPixmap(
                vistrails.core.system.vistrails_root_directory() +
                '/gui/resources/images/multiline_string_icon.png'))

        # icons for the port list combination modes
        self.DOT_PRODUCT_ICON = QtGui.QIcon(QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/macro.png'))

        self.CROSS_PRODUCT_ICON = QtGui.QIcon(QtGui.QPixmap(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/remove_param.png'))

        #### COLORS ####
        # Color for the PIP frame
        self.PIP_FRAME_COLOR = QtGui.QColor(
            *(ColorByName.get_int('yellow_light')))

        # Color for invalid parameter frames
        self.PARAM_INVALID_COLOR = QtGui.QColor('#efef00')

        # Color of selected methods in the modules method area
        self.METHOD_SELECT_COLOR = QtGui.QColor(
            *ColorByName.get_int('yellow_light'))

        # Color of the hover/unhover alias labels
        self.HOVER_DEFAULT_COLOR = QtGui.QColor(
            *ColorByName.get_int('black'))
        self.HOVER_SELECT_COLOR = QtGui.QColor(
            *ColorByName.get_int('blue'))
        
        # colors for debug messages
        self.DEBUG_COLORS = {
                'DEBUG': QtGui.QColor("#777777"),
                'INFO': QtGui.QColor(QtCore.Qt.black),
                'WARNING': QtGui.QColor("#707000"),
                'CRITICAL': QtGui.QColor(QtCore.Qt.red),
            }
        class QTransparentColor(QtGui.QColor):
            def name(self):
                return 'transparent'
        self.DEBUG_FILTER_BACKGROUND_COLOR = QTransparentColor("transparent")
        
class MacTheme(DefaultTheme):
    def __init__(self):
        
        DefaultTheme.__init__(self)
        #### ICONS & IMAGES ####
        #The application icon
        self.APPLICATION_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/vistrails_icon.png')

        #The application pixmap
        self.APPLICATION_PIXMAP = QtGui.QPixmap(
             vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/vistrails_icon.png')

        # Toolbar icon for close button on the vistrail tabbar
        self.VIEW_MANAGER_CLOSE_ICON = QtGui.QIcon(
            vistrails.core.system.vistrails_root_directory() +
            '/gui/resources/images/mac/closetab.png')
        
        #### FONTS ####
        # Font for PythonSource
        self.PYTHON_SOURCE_EDITOR_FONT = QtGui.QFont('Monaco', 11, 
                                                     QtGui.QFont.Normal)
        
        # Font for Splash Screen messages
        self.SPLASH_SCREEN_FONT = QtGui.QFont('Helvetica', 10,
                                              QtGui.QFont.Light)
        
class LinuxTheme(DefaultTheme):
    def __init__(self):
        DefaultTheme.__init__(self)
        #### FONTS ####
        # Font for PythonSource
        self.PYTHON_SOURCE_EDITOR_FONT = QtGui.QFont('Monospace', 10, 
                                                     QtGui.QFont.Normal)
                
class ThemeHolder(object):
    """
    ThemeHolder is a class holding a theme and exposed that theme
    atrributes as attributes of itself. This is useful for global
    import of the CurrentTheme variable
    
    """
    def __init__(self):
        object.__init__(self)
        self.theme = None

    def setTheme(self, theme):
        """ setTheme(theme: subclass of DefaultTheme) -> None
        Set the current theme to theme
        
        """
        # This way, the lookups into the theme are much faster, since
        # there's no function calls involved
        self.__dict__.update(theme.__dict__)

def get_current_theme():
    """get_current_theme() -> subclass of DefaultTheme
    Instantiates the theme according to the current platform """
    if vistrails.core.system.systemType in ['Darwin']:
        return MacTheme()
    elif vistrails.core.system.systemType in ['Linux']:
        return LinuxTheme()
    else:
        return DefaultTheme()
    
def initializeCurrentTheme():
    """ initializeCurrentTheme() -> None
    Assign the current theme to the default theme
    
    """
    CurrentTheme.setTheme(get_current_theme())

global CurrentTheme
CurrentTheme = ThemeHolder()
    
################################################################################


class TestPresetColor(unittest.TestCase):
    """
    A few simple tests to make sure Preset is working as expected
    
    """
    def setUp(self):
        vistrails.gui.qt.createBogusQtGuiApp()

    def testColorValues(self):
        initializeCurrentTheme()
        self.assertEquals(CurrentTheme.CONNECTION_CONTROL_POINTS,
                          20)
        
if __name__ == '__main__':
    unittest.main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""This file contains classes working with cell helper widgets, i.e. toolbar,
esizer, etc.:
  QCellWidget
  QCellToolBar
"""

from __future__ import division

import datetime
import os
from PyQt4 import QtCore, QtGui
import tempfile

from vistrails.core import debug
from vistrails.core.modules.output_modules import FileMode
from vistrails.core.system import strftime

from .analogy_api import SpreadsheetAnalogy
from .spreadsheet_config import configuration
import cell_rc
import celltoolbar_rc


class QCellWidget(QtGui.QWidget):
    """
    QCellWidget is the base cell class. All types of spreadsheet cells
    should inherit from this.

    """
    save_formats = ["Images (*.png *.xpm *.jpg)",
                    "Portable Document Format (*.pdf)"]

    def __init__(self, parent=None, flags=QtCore.Qt.WindowFlags()):
        """ QCellWidget(parent: QWidget) -> QCellWidget
        Instantiate the cell and helper properties

        """
        QtGui.QWidget.__init__(self, parent, flags)
        self._historyImages = []
        self._player = QtGui.QLabel(self.parent())
        self._player.setAutoFillBackground(True)
        self._player.setFocusPolicy(QtCore.Qt.NoFocus)
        self._player.setScaledContents(True)
        self._playerTimer = QtCore.QTimer()
        self._playerTimer.setSingleShot(True)
        self._currentFrame = 0
        self._playing = False
        # cell can be captured if it re-implements saveToPNG
        self._capturingEnabled = (not isinstance(self, QCellWidget) and
                                  hasattr(self, 'saveToPNG'))
        self._output_module = None
        self._output_configuration = None
        self.connect(self._playerTimer,
                     QtCore.SIGNAL('timeout()'),
                     self.playNextFrame)
        if configuration.fixedCellSize:
            self.setSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
            self.setFixedSize(200, 180)

    def setAnimationEnabled(self, enabled):
        """ setAnimationEnabled(enabled: bool) -> None

        """
        self._capturingEnabled = enabled
        if not enabled:
            self.clearHistory()

    def saveToPNG(self, filename):
        """ saveToPNG(filename: str) -> Bool
        Abtract function for saving the current widget contents to an
        image file
        Returns True when succesful

        """
        debug.critical('saveToPNG() is unimplemented by the inherited cell')

    def saveToHistory(self):
        """ saveToHistory() -> None
        Save the current contents to the history

        """
        # Generate filename
        current = datetime.datetime.now()
        tmpDir = tempfile.gettempdir()
        fn = ( "hist_" + strftime(current, "%Y_%m_%d__%H_%M_%S") +
               "_" + str(current.microsecond)+".png")
        fn = os.path.join(tmpDir, fn)
        if self.saveToPNG(fn):
            self._historyImages.append(fn)

    def clearHistory(self):
        """ clearHistory() -> None
        Clear all history files

        """
        for fn in self._historyImages:
            os.remove(fn)
        self._historyImages = []

    def deleteLater(self):
        """ deleteLater() -> None
        Make sure to clear history and delete the widget

        """
        self.clearHistory()
        QtGui.QWidget.deleteLater(self)

    def updateContents(self, inputPorts):
        """ updateContents(inputPorts: tuple)
        Make sure to capture to history

        """
        # Capture window into history for playback
        if self._capturingEnabled:
            self.saveToHistory()

    def resizeEvent(self, e):
        """ resizeEvent(e: QEvent) -> None
        Re-adjust the player widget

        """
        QtGui.QWidget.resizeEvent(self, e)

        if self._player.isVisible():
            self._player.setGeometry(self.geometry())

    def setPlayerFrame(self, frame):
        """ setPlayerFrame(frame: int) -> None
        Set the player to display a particular frame number

        """
        if (len(self._historyImages)==0):
            return
        if frame>=len(self._historyImages):
            frame = frame % len(self._historyImages)
        if frame>=len(self._historyImages):
            return
        self._player.setPixmap(QtGui.QPixmap(self._historyImages[frame]))

    def startPlayer(self):
        """ startPlayer() -> None
        Adjust the size of the player to the cell and show it

        """
        if not self._capturingEnabled:
            return
        self._player.setParent(self.parent())
        self._player.setGeometry(self.geometry())
        self._player.raise_()
        self._currentFrame = -1
        self.playNextFrame()
        self._player.show()
        self.hide()
        self._playing = True

    def stopPlayer(self):
        """ startPlayer() -> None
        Adjust the size of the player to the cell and show it

        """
        if not self._capturingEnabled:
            return
        self._playerTimer.stop()
        self._player.hide()
        self.show()
        self._playing = False

    def showNextFrame(self):
        """ showNextFrame() -> None
        Display the next frame in the history

        """
        self._currentFrame += 1
        if self._currentFrame>=len(self._historyImages):
            self._currentFrame = 0
        self.setPlayerFrame(self._currentFrame)

    def playNextFrame(self):
        """ playNextFrame() -> None
        Display the next frame in the history and start the timer for
        the frame after

        """
        self.showNextFrame()
        self._playerTimer.start(100)

    def grabWindowPixmap(self):
        """ grabWindowPixmap() -> QPixmap
        Widget special grabbing function

        """
        return QtGui.QPixmap.grabWidget(self)

    def dumpToFile(self, filename):
        """ dumpToFile(filename: str, dump_as_pdf: bool) -> None
        Dumps itself as an image to a file, calling grabWindowPixmap """
        pixmap = self.grabWindowPixmap()
        ext = os.path.splitext(filename)[1].lower()
        if not ext:
            pixmap.save(filename, 'PNG')
        elif ext == '.pdf':
            self.saveToPDF(filename)
        else:
            pixmap.save(filename)

    def saveToPDF(self, filename):
        printer = QtGui.QPrinter()

        printer.setOutputFormat(QtGui.QPrinter.PdfFormat)
        printer.setOutputFileName(filename)
        pixmap = self.grabWindowPixmap()
        size = pixmap.size()
        printer.setPaperSize(QtCore.QSizeF(size.width(), size.height()),
                             QtGui.QPrinter.Point)
        painter = QtGui.QPainter()
        painter.begin(printer)
        rect = painter.viewport()
        size.scale(rect.size(), QtCore.Qt.KeepAspectRatio)
        painter.setViewport(rect.x(), rect.y(), size.width(), size.height())
        painter.setWindow(pixmap.rect())
        painter.drawPixmap(0, 0, pixmap)
        painter.end()

    def set_output_module(self, output_module, configuration=None):
        self._output_module = output_module
        self._output_configuration = configuration

    def has_file_output_mode(self):
        # from vistrails.core.modules.output_modules import FileMode
        if self._output_module is None:
            return False
        for mode in self._output_module.get_sorted_mode_list():
            if issubclass(mode, FileMode):
                return True
        return False

    def get_file_output_modes(self):
        modes = []
        if self._output_module is not None:
            for mode_cls in self._output_module.get_sorted_mode_list():
                if issubclass(mode_cls, FileMode):
                    modes.append(mode_cls)
        return modes

    def get_conf_file_format(self):
        if (self._output_configuration is not None and
            'format' in self._output_configuration):
            return self._output_configuration['format']
        return None

    def save_via_file_output(self, filename, mode_cls, save_format=None):
        mode = mode_cls()
        mode_config = self._output_module.get_mode_config(mode)
        mode_config['file'] = filename
        if save_format is not None:
            mode_config['format'] = save_format
        mode.compute_output(self._output_module, mode_config)


class QCellToolBar(QtGui.QToolBar):
    """
    CellToolBar is inherited from QToolBar with some functionalities
    for interacting with CellHelpers

    """
    def __init__(self, sheet):
        """ CellToolBar(sheet: SpreadsheetSheet) -> CellToolBar
        Initialize the cell toolbar by calling the user-defined
        toolbar construction function

        """
        QtGui.QToolBar.__init__(self,sheet)
        self.setOrientation(QtCore.Qt.Horizontal)
        self.sheet = sheet
        self.row = -1
        self.col = -1
        self.layout().setMargin(0)
        self.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Preferred)
        pixmap = self.style().standardPixmap(QtGui.QStyle.SP_DialogCloseButton)
        self.addSaveCellAction()
        self.addExecuteCellAction()
        self.appendAction(QCellToolBarRemoveCell(QtGui.QIcon(pixmap), self))
        self.appendAction(QCellToolBarMergeCells(QtGui.QIcon(':celltoolbar/mergecells.png'), self))
        self.createToolBar()

    def addAnimationButtons(self):
        """ addAnimationButtons() -> None

        """
        self.appendAction(QCellToolBarCaptureToHistory(self))
        self.appendAction(QCellToolBarPlayHistory(self))
        self.appendAction(QCellToolBarClearHistory(self))

    def addSaveCellAction(self):
        if not hasattr(self, 'saveActionVar'):
            self.saveActionVar = QCellToolBarSelectedCell(
                    QtGui.QIcon(":/images/camera.png"),
                    "Save cell",
                    self)
            self.saveActionVar.setStatusTip("Export this cell only")

            self.connect(self.saveActionVar, QtCore.SIGNAL('triggered(bool)'),
                         self.exportCell)
        self.appendAction(self.saveActionVar)

    def exportCell(self, checked=False):
        cell = self.sheet.getCell(self.row, self.col)
        if cell.has_file_output_mode():
            modes = cell.get_file_output_modes()
            formats = []
            format_map = {}
            for mode in modes:
                for m_format in mode.get_formats():
                    if m_format not in format_map:
                        formats.append(m_format)
                        format_map[m_format] = mode
            selected_filter = None
            if cell.get_conf_file_format() is not None:
                selected_filter = '(*.%s)' % cell.get_conf_file_format()
            (filename, save_format) = \
                    QtGui.QFileDialog.getSaveFileNameAndFilter(
                        self, "Select a File to Export the Cell",
                        ".", ';;'.join(['(*.%s)' % f for f in formats]),
                        selected_filter)
            if filename:
                save_mode = format_map[save_format[3:-1]]
                cell.save_via_file_output(filename, save_mode)
        else:
            if not cell.save_formats:
                QtGui.QMessageBox.information(
                        self, "Export cell",
                        "This cell type doesn't provide any export option")
                return
            filename = QtGui.QFileDialog.getSaveFileName(
                self, "Select a File to Export the Cell",
                ".", ';;'.join(cell.save_formats))
            if filename:
                cell.dumpToFile(filename)

    def addExecuteCellAction(self):
        if not hasattr(self, 'executeActionVar'):
            self.executeActionVar = QCellToolBarSelectedCell(
                    QtGui.QIcon(":/images/view-refresh.png"),
                    "Re-execute cell",
                    self)
            self.executeActionVar.setStatusTip("Re-execute this cell")

            self.connect(self.executeActionVar, QtCore.SIGNAL('triggered(bool)'),
                         self.executeCell)
        self.appendAction(self.executeActionVar)

    def executeCell(self, checked=False):
        from spreadsheet_execute import executePipelineWithProgress
        #cell = self.sheet.getCell(self.row, self.col)
        info = self.sheet.getCellPipelineInfo(self.row, self.col)
        if info:
            info = info[0]
            mId = info['moduleId']
            pipeline = self.sheet.setPipelineToLocateAt(self.row, self.col,
                                                        info['pipeline'],
                                                        [mId])
            executePipelineWithProgress(pipeline, 'Re-execute Cell',
                                        current_version=info['version'],
                                        actions=info['actions'],
                                        reason=info['reason'],
                                        locator=info['locator'],
                                        controller=info['controller'],
                                        sinks=[mId])



    def createToolBar(self):
        """ createToolBar() -> None
        A user-defined method for customizing the toolbar. This is
        going to be an empty method here for inherited classes to
        override.

        """
        self.addAnimationButtons()

    def snapTo(self, row, col):
        """ snapTo(row, col) -> None
        Assign which row and column the toolbar should be snapped to

        """
        self.row = row
        self.col = col
        self.updateToolBar()

    def updateToolBar(self):
        """ updateToolBar() -> None
        This will get called when the toolbar widgets need to have
        their status updated. It sends out needUpdateStatus signal
        to let the widget have a change to update their own status

        """
        cellWidget = self.sheet.getCell(self.row, self.col)
        for action in self.actions():
            action.emit(QtCore.SIGNAL('needUpdateStatus'),
                        (self.sheet, self.row, self.col, cellWidget))

    def connectAction(self, action, widget):
        """ connectAction(action: QAction, widget: QWidget) -> None
        Connect actions to special slots of a widget

        """
        if hasattr(widget, 'updateStatus'):
            self.connect(action, QtCore.SIGNAL('needUpdateStatus'),
                         widget.updateStatus)
        if hasattr(widget, 'triggeredSlot'):
            self.connect(action, QtCore.SIGNAL('triggered()'),
                         widget.triggeredSlot)
        if hasattr(widget, 'toggledSlot'):
            self.connect(action, QtCore.SIGNAL('toggled(bool)'),
                         widget.toggledSlot)

    def appendAction(self, action):
        """ appendAction(action: QAction) -> QAction
        Setup and add action to the tool bar

        """
        action.toolBar = self
        self.addAction(action)
        self.connectAction(action, action)
        return action

    def appendWidget(self, widget):
        """ appendWidget(widget: QWidget) -> QAction
        Setup the widget as an action and add it to the tool bar

        """
        action = self.addWidget(widget)
        widget.toolBar = self
        action.toolBar = self
        self.connectAction(action, widget)
        return action

    def getSnappedWidget(self):
        """ getSnappedWidget() -> QWidget
        Return the widget being snapped by the toolbar

        """
        if self.row>=0 and self.col>=0:
            return self.sheet.getCell(self.row, self.col)
        else:
            return None


class QCellToolBarSelectedCell(QtGui.QAction):
    """
    QCellToolBarSelectedCell is an action only visible if the cell isn't empty.
    """
    def updateStatus(self, info):
        """ updateStatus(info: tuple) -> None
        Updates the status of the button based on the input info

        """
        (sheet, row, col, cellWidget) = info
        self.setVisible(cellWidget is not None)


class QCellToolBarRemoveCell(QCellToolBarSelectedCell):
    """
    QCellToolBarRemoveCell is the action to clear the current cell

    """
    def __init__(self, icon, parent=None):
        """ QCellToolBarRemoveCell(icon: QIcon, parent: QWidget)
                                   -> QCellToolBarRemoveCell
        Setup the image, status tip, etc. of the action

        """
        QCellToolBarSelectedCell.__init__(self,
                                          icon,
                                          "&Clear the current cell",
                                          parent)
        self.setStatusTip("Clear the current cell")

    def triggeredSlot(self, checked=False):
        """ toggledSlot(checked: boolean) -> None
        Execute the action when the button is clicked

        """
        cellWidget = self.toolBar.getSnappedWidget()
        r = QtGui.QMessageBox.question(cellWidget, 'Clear cell',
                                       'Are you sure to clear the cell?',
                                       QtGui.QMessageBox.Yes |
                                       QtGui.QMessageBox.No,
                                       QtGui.QMessageBox.No)
        if (r==QtGui.QMessageBox.Yes):
            self.toolBar.sheet.deleteCell(self.toolBar.row, self.toolBar.col)


class QCellToolBarMergeCells(QtGui.QAction):
    """
    QCellToolBarMergeCells is the action to merge selected cells to a
    single cell if they are in consecutive poisitions

    """
    def __init__(self, icon, parent=None):
        """ QCellToolBarMergeCells(icon: QIcon, parent: QWidget)
                                   -> QCellToolBarMergeCells
        Setup the image, status tip, etc. of the action

        """
        QtGui.QAction.__init__(self,
                               icon,
                               "&Merge cells",
                               parent)
        self.setStatusTip("Merge selected cells to a single cell if "
                          "they are in consecutive poisitions")
        self.setCheckable(True)

    def triggeredSlot(self):
        """ toggledSlot() -> None
        Execute the action when the button is clicked

        """
        # Merge
        if self.isChecked():
            sheet = self.toolBar.sheet
            selectedCells = sorted(sheet.getSelectedLocations())
            topLeft = selectedCells[0]
            bottomRight = selectedCells[-1]
            sheet.setSpan(topLeft[0], topLeft[1],
                          bottomRight[0]-topLeft[0]+1,
                          bottomRight[1]-topLeft[1]+1)
        else:
            sheet = self.toolBar.sheet
            selectedCells = sorted(sheet.getSelectedLocations())
            for (row, col) in selectedCells:
                sheet.setSpan(row, col, 1, 1)
        sheet.clearSelection()
        self.toolBar.updateToolBar()

    def updateStatus(self, info):
        """ updateStatus(info: tuple) -> None
        Updates the status of the button based on the input info

        """
        (sheet, row, col, cellWidget) = info
        selectedCells = sorted(sheet.getSelectedLocations())

        # Will not show up if there is no cell selected
        if len(selectedCells)==0:
            self.setVisible(False)

        # If there is a single cell selected, only show up if it has
        # been merged before so that user can un-merge cells
        elif len(selectedCells)==1:
            showUp = False
            if selectedCells[0]==(row, col):
                span = sheet.getSpan(row, col)
                if span[0]>1 or span[1]>1:
                    showUp = True
            if showUp:
                self.setChecked(True)
                self.setVisible(True)
            else:
                self.setVisible(False)

        # If there are multiple cells selected, only show up if they
        # can be merged, i.e. cells are in consecutive position and
        # none of them is already merged
        else:
            showUp = False
            validRange = False
            topLeft = selectedCells[0]
            bottomRight = selectedCells[-1]
            fullCount = (bottomRight[0]-topLeft[0]+1)*(bottomRight[1]-topLeft[1]+1)
            validRange = len(selectedCells)==fullCount
            if validRange:
                showUp = True
                for (r, c) in selectedCells:
                    span = sheet.getSpan(r, c)
                    if span[0]>1 or span[1]>1:
                        showUp = False
                        break
            if showUp:
                self.setChecked(False)
                self.setVisible(True)
            else:
                self.setVisible(False)


class QCellToolBarCaptureToHistory(QtGui.QAction):
    """
    QCellToolBarCaptureToHistory is the action to capture the
    underlying widget to history for play back. The cell type must
    support function saveToPNG(filename)

    """
    def __init__(self, parent=None):
        """ QCellToolBarCaptureToHistory(parent: QWidget)
                                         -> QCellToolBarCaptureToHistory
        Setup the image, status tip, etc. of the action

        """
        QtGui.QAction.__init__(self,
                               QtGui.QIcon(":/images/camera_mount.png"),
                               "&Capture image to history",
                               parent)
        self.setStatusTip("Capture the cell contents to the history for "
                          "playback later")

    def triggeredSlot(self, checked=False):
        """ toggledSlot(checked: boolean) -> None
        Execute the action when the button is clicked

        """
        cellWidget = self.toolBar.getSnappedWidget()
        self.toolBar.hide()
        cellWidget.saveToHistory()
        self.toolBar.updateToolBar()
        self.toolBar.show()

    def updateStatus(self, info):
        """ updateStatus(info: tuple) -> None
        Updates the status of the button based on the input info

        """
        (sheet, row, col, cellWidget) = info
        if cellWidget:
            self.setVisible(cellWidget._capturingEnabled)
        else:
            self.setVisible(False)


class QCellToolBarPlayHistory(QtGui.QAction):
    """
    QCellToolBarPlayHistory is the action to play the history as an
    animation

    """
    def __init__(self, parent=None):
        """ QCellToolBarPlayHistory(parent: QWidget)
                                    -> QCellToolBarPlayHistory
        Setup the image, status tip, etc. of the action

        """
        self.icons = [QtGui.QIcon(":/images/player_play.png"),
                      QtGui.QIcon(":/images/player_pause.png")]
        self.toolTips = ["&Play the history",
                         "Pa&use the history playback"]
        self.statusTips = ["Playback all image files kept in the history",
                           "Pause the playback"]
        QtGui.QAction.__init__(self, self.icons[0], self.toolTips[0], parent)
        self.setStatusTip(self.statusTips[0])
        self.status = 0

    def triggeredSlot(self, checked=False):
        """ toggledSlot(checked: boolean) -> None
        Execute the action when the button is clicked

        """
        cellWidget = self.toolBar.getSnappedWidget()
        if self.status==0:
            cellWidget.startPlayer()
        else:
            cellWidget.stopPlayer()
        self.toolBar.updateToolBar()

    def updateStatus(self, info):
        """ updateStatus(info: tuple) -> None
        Updates the status of the button based on the input info

        """
        (sheet, row, col, cellWidget) = info
        if cellWidget:
            self.setVisible(cellWidget._capturingEnabled)
            newStatus = int(cellWidget._playing)
            if newStatus!=self.status:
                self.status = newStatus
                self.setIcon(self.icons[self.status])
                self.setToolTip(self.toolTips[self.status])
                self.setStatusTip(self.statusTips[self.status])
            self.setEnabled(len(cellWidget._historyImages)>0)
        else:
            self.setVisible(False)


class QCellToolBarClearHistory(QtGui.QAction):
    """
    QCellToolBarClearHistory is the action to reset cell history

    """
    def __init__(self, parent=None):
        """ QCellToolBarClearHistory(parent: QWidget)
                                     -> QCellToolBarClearHistory
        Setup the image, status tip, etc. of the action

        """
        QtGui.QAction.__init__(self,
                               QtGui.QIcon(":/images/noatunloopsong.png"),
                               "&Clear this cell history",
                               parent)
        self.setStatusTip("Clear the cell history and its temporary "
                          "image files on disk")

    def triggeredSlot(self, checked=False):
        """ toggledSlot(checked: boolean) -> None
        Execute the action when the button is clicked

        """
        cellWidget = self.toolBar.getSnappedWidget()
        cellWidget.clearHistory()
        self.toolBar.updateToolBar()

    def updateStatus(self, info):
        """ updateStatus(info: tuple) -> None
        Updates the status of the button based on the input info

        """
        (sheet, row, col, cellWidget) = info
        if cellWidget:
            self.setVisible(cellWidget._capturingEnabled)
            self.setEnabled((len(cellWidget._historyImages)>0
                             and cellWidget._playing==False))
        else:
            self.setVisible(False)


class QCellContainer(QtGui.QWidget):
    """ QCellContainer is a simple QWidget containing the actual cell
    widget as a child. This also acts as a sentinel protecting the
    actual cell widget from being destroyed by sheet widgets
    (e.g. QTableWidget) where they take control of the cell widget.

    """
    def __init__(self, widget=None, parent=None):
        """ QCellContainer(parent: QWidget) -> QCellContainer
        Create an empty container

        """
        QtGui.QWidget.__init__(self, parent)
        layout = QtGui.QVBoxLayout()
        layout.setSpacing(0)
        layout.setMargin(0)
        self.setLayout(layout)
        self.containedWidget = None
        self.setWidget(widget)
        self.toolBar = None

    def setWidget(self, widget):
        """ setWidget(widget: QWidget) -> None
        Set the contained widget of this container

        """
        if widget!=self.containedWidget:
            if self.containedWidget:
                self.layout().removeWidget(self.containedWidget)
                self.containedWidget.deleteLater()
                self.toolBar = None
            if widget:
                widget.setParent(self)
                self.layout().addWidget(widget)
                widget.show()
            self.containedWidget = widget

    def widget(self):
        """ widget() -> QWidget
        Return the contained widget

        """
        return self.containedWidget

    def takeWidget(self):
        """ widget() -> QWidget
        Take the contained widget out without deleting

        """
        widget = self.containedWidget
        if self.containedWidget:
            self.layout().removeWidget(self.containedWidget)
            self.containedWidget.setParent(None)
            self.containedWidget = None
        self.toolBar = None
        return widget


class QCellPresenter(QtGui.QLabel):
    """
    QCellPresenter represents a cell in the Editing Mode. It has an
    info bar on top and control dragable icons on the bottom

    """
    def __init__(self, parent=None):
        """ QCellPresenter(parent: QWidget) -> QCellPresenter
        Create the layout of the widget

        """
        QtGui.QLabel.__init__(self, parent)
        self.setAutoFillBackground(True)
        self.setScaledContents(True)
        self.setMargin(0)
        self.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.cellWidget = None

        layout = QtGui.QGridLayout(self)
        layout.setSpacing(2)
        layout.setMargin(self.margin())
        layout.setRowStretch(1, 1)
        self.setLayout(layout)

        self.info = QPipelineInfo()
        layout.addWidget(self.info, 0, 0, 1, 2)

        self.manipulator = QCellManipulator()
        layout.addWidget(self.manipulator, 1, 0, 1, 2)

    def assignCellWidget(self, cellWidget):
        """ updateFromCellWidget(cellWidget: QWidget) -> None
        Assign a cell widget to this presenter

        """
        self.cellWidget = cellWidget
        if cellWidget:
            if hasattr(cellWidget, 'grabWindowPixmap'):
                bgPixmap = cellWidget.grabWindowPixmap()
            else:
                bgPixmap = QtGui.QPixmap.grabWidget(cellWidget)
            self.info.show()
        else:
            self.info.hide()
            bgPixmap = QtGui.QPixmap.grabWidget(self)
        self.thumbnail = QtGui.QPixmap(bgPixmap)
        painter = QtGui.QPainter(bgPixmap)
        painter.fillRect(bgPixmap.rect(),
                         QtGui.QBrush(QtGui.QColor(175, 198, 229, 196)))
        painter.end()
        self.setPixmap(bgPixmap)

    def assignCell(self, sheet, row, col):
        """ assignCell(sheet: Sheet, row: int, col: int) -> None
        Assign a sheet cell to the presenter

        """
        self.manipulator.assignCell(sheet, row, col)
        self.assignCellWidget(sheet.getCell(row, col))
        info = sheet.getCellPipelineInfo(row, col)
        self.info.updateInfo(info)

    def releaseCellWidget(self):
        """ releaseCellWidget() -> QWidget
        Return the ownership of self.cellWidget to the caller

        """
        cellWidget = self.cellWidget
        self.assignCellWidget(None)
        self.manipulator.assignCell(None, -1, -1)
        if cellWidget:
            cellWidget.setParent(None)
        return cellWidget

    def deleteLater(self):
        """ deleteLater() -> None
        Make sure to delete the cell widget if it exists

        """
        if (self.cellWidget):
            self.cellWidget.deleteLater()
        QtGui.QLabel.deleteLater(self)


class QInfoLineEdit(QtGui.QLineEdit):
    """
    QInfoLineEdit is wrapper for a transparent, un-frame, read-only
    line edit

    """
    def __init__(self, parent=None):
        """ QInfoLineEdit(parent: QWidget) -> QInfoLineEdit
        Initialize the line edit

        """
        QtGui.QLineEdit.__init__(self, parent)
        self.setReadOnly(True)
        self.setFrame(False)
        pal = QtGui.QPalette(self.palette())
        pal.setBrush(QtGui.QPalette.Base,
                     QtGui.QBrush(QtCore.Qt.NoBrush))
        self.setPalette(pal)


class QInfoLabel(QtGui.QLabel):
    """
    QInfoLabel is wrapper for a transparent, bolded label

    """
    def __init__(self, text='', parent=None):
        """ QInfoLabel(text: str, parent: QWidget) -> QInfoLabel
        Initialize the line edit

        """
        QtGui.QLabel.__init__(self, text, parent)
        font = QtGui.QFont(self.font())
        font.setBold(True)
        self.setFont(font)


class QPipelineInfo(QtGui.QFrame):
    """
    QPipelineInfo displays information about the executed pipeline of
    a cell. It has 3 static lines: Vistrail name, (pipeline name,
    pipeline id) and the cell type

    """
    def __init__(self, parent=None):
        """ QPipelineInfo(parent: QWidget) -> None
        Create the 3 information lines

        """
        QtGui.QFrame.__init__(self, parent)
        self.setAutoFillBackground(True)
        self.setFrameStyle(QtGui.QFrame.NoFrame)
        self.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Maximum)

        pal = QtGui.QPalette(self.palette())
        color = QtGui.QColor(pal.brush(QtGui.QPalette.Base).color())
        color.setAlpha(196)
        pal.setBrush(QtGui.QPalette.Base, QtGui.QBrush(color))
        self.setPalette(pal)

        topLayout = QtGui.QVBoxLayout(self)
        topLayout.setSpacing(0)
        topLayout.setMargin(0)
        self.setLayout(topLayout)

        hLine = QtGui.QFrame()
        hLine.setFrameStyle(QtGui.QFrame.HLine | QtGui.QFrame.Plain)
        hLine.setFixedHeight(1)
        topLayout.addWidget(hLine)

        layout = QtGui.QGridLayout()
        layout.setSpacing(2)
        layout.setMargin(2)
        topLayout.addLayout(layout)

        self.edits = []
        texts = ['Vistrail', 'Index', 'Created by']
        for i in xrange(len(texts)):
            label = QInfoLabel(texts[i])
            layout.addWidget(label, i, 0, 1, 1)
            edit = QInfoLineEdit()
            self.edits.append(edit)
            layout.addWidget(edit, i, 1, 1, 1)

        topLayout.addStretch()
        hLine = QtGui.QFrame()
        hLine.setFrameStyle(QtGui.QFrame.HLine | QtGui.QFrame.Plain)
        hLine.setFixedHeight(1)
        topLayout.addWidget(hLine)

    def updateInfo(self, info):
        """ updateInfo(info: (dict, pid)) -> None
        Update the information of a pipeline info

        """
        if info!=None and info[0]['locator']!=None:
            self.edits[0].setText(str(info[0]['locator'].name))
            self.edits[1].setText('(Pipeline: %d, Module: %d)'
                                  % (info[0]['version'], info[0]['moduleId']))
            self.edits[2].setText(str(info[0]['reason']))
        else:
            for edit in self.edits:
                edit.setText('N/A')


class QCellManipulator(QtGui.QFrame):
    """
    QCellManipulator contains several dragable icons that allow users
    to move/copy or perform some operation from one cell to
    another. It also inclues a button for update the pipeline under
    the cell to be a new version on the pipeline. It is useful for the
    parameter exploration talks back to the builder

    """
    def __init__(self, parent=None):
        """ QPipelineInfo(parent: QWidget) -> None
        Create the 3 information lines

        """
        QtGui.QFrame.__init__(self, parent)
        self.setAcceptDrops(True)
        self.setFrameStyle(QtGui.QFrame.NoFrame)
        self.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)

        layout = QtGui.QVBoxLayout(self)
        layout.setSpacing(0)
        layout.setMargin(0)
        self.setLayout(layout)

        layout.addStretch()

        bLayout = QtGui.QHBoxLayout()
        layout.addLayout(bLayout)

        bInfo = [(':/images/copy_cell.png',
                  'Drag to copy this cell to another place',
                  'copy', 'Copy'),
                 (':/images/move_cell.png',
                  'Drag to move this cell to another place',
                  'move', 'Move'),
                 (':/images/create_analogy.png',
                  'Drag to create an analogy between this cell and another one',
                  'create_analogy', 'Create\nAnalogy'),
                 (':/images/apply_analogy.png',
                  'Drag to apply the current analogy to this cell and put it '
                  'at another place', 'apply_analogy', 'Apply\nAnalogy')]

        self.buttons = []

        bLayout.addStretch()
        for b in bInfo:
            button = QCellDragLabel(QtGui.QPixmap(b[0]))
            button.setToolTip(b[1])
            button.setStatusTip(b[1])
            button.action = b[2]
            vLayout = QtGui.QVBoxLayout()
            vLayout.addWidget(button)
            label = QtGui.QLabel(b[3])
            label.setAlignment(QtCore.Qt.AlignCenter)
            vLayout.addWidget(label)
            bLayout.addLayout(vLayout)
            self.buttons.append(button)
            self.buttons.append(label)

        self.updateButton = QtGui.QToolButton()
        self.updateButton.setIconSize(QtCore.QSize(64, 64))
        self.updateButton.setIcon(QtGui.QIcon(QtGui.QPixmap(
            ':/images/update.png')))
        self.updateButton.setAutoRaise(True)
        self.updateButton.setToolTip('Add this cell as a new version')
        self.updateButton.setStatusTip(self.updateButton.toolTip())
        self.updateButton.setText('Create Version')
        self.updateButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
        self.connect(self.updateButton, QtCore.SIGNAL('clicked(bool)'),
                     self.updateVersion)
        self.buttons.append(self.updateButton)

        self.locateButton = QtGui.QToolButton()
        self.locateButton.setIconSize(QtCore.QSize(64, 64))
        self.locateButton.setIcon(QtGui.QIcon(QtGui.QPixmap(
            ':/images/locate.png')))
        self.locateButton.setAutoRaise(True)
        self.locateButton.setToolTip('Select this pipeline in the version tree ')
        self.locateButton.setStatusTip(self.locateButton.toolTip())
        self.locateButton.setText('Locate Version')
        self.locateButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
        self.connect(self.locateButton, QtCore.SIGNAL('clicked(bool)'),
                     self.locateVersion)
        self.buttons.append(self.locateButton)


        uLayout = QtGui.QHBoxLayout()
        uLayout.addStretch()
        uLayout.addWidget(self.locateButton)
        uLayout.addWidget(self.updateButton)
        uLayout.addStretch()
        layout.addLayout(uLayout)

        bLayout.addStretch()

        layout.addStretch()

        self.innerRubberBand = QtGui.QRubberBand(QtGui.QRubberBand.Rectangle,
                                                 self)

    def assignCell(self, sheet, row, col):
        """ assignCell(sheet: Sheet, row: int, col: int) -> None
        Assign a cell to the manipulator, so it knows where to drag
        and drop

        """
        self.cellInfo = (sheet, row, col)
        for b in self.buttons:
            if hasattr(b, 'updateCellInfo'):
                b.updateCellInfo(self.cellInfo)
            if sheet and sheet.getCell(row, col)!=None:
                widget = sheet.getCell(row, col)
                b.setVisible(not isinstance(widget, QCellPresenter) or
                             widget.cellWidget is not None)
            else:
                b.setVisible(False)
        self.updateButton.setVisible(False)
        if sheet:
            info = sheet.getCellPipelineInfo(row, col)
            if info:
                self.updateButton.setVisible(len(info[0]['actions'])>0)

    def dragEnterEvent(self, event):
        """ dragEnterEvent(event: QDragEnterEvent) -> None
        Set to accept drops from the other cell info

        """
        mimeData = event.mimeData()
        if hasattr(mimeData, 'cellInfo'):
            if (mimeData.cellInfo==self.cellInfo or
                mimeData.cellInfo[0]==None or
                self.cellInfo[0]==None):
                event.ignore()
            else:
                event.setDropAction(QtCore.Qt.MoveAction)
                event.accept()
                self.highlight()
        else:
            event.ignore()

    def dragLeaveEvent(self, event):
        """ dragLeaveEvent(event: QDragLeaveEvent) -> None
        Unhighlight when the cursor leaves

        """
        self.highlight(False)

    def dropEvent(self, event):
        """ dragLeaveEvent(event: QDragLeaveEvent) -> None
        Unhighlight when the cursor leaves

        """
        self.highlight(False)
        mimeData = event.mimeData()
        action = mimeData.action
        cellInfo = mimeData.cellInfo
        manipulator = mimeData.manipulator
        if action in ['move', 'copy', 'create_analogy', 'apply_analogy']:
            event.setDropAction(QtCore.Qt.MoveAction)
            event.accept()

            if action=='move':
                self.cellInfo[0].swapCell(self.cellInfo[1], self.cellInfo[2],
                                          cellInfo[0], cellInfo[1], cellInfo[2])
                manipulator.assignCell(*self.cellInfo)
                self.assignCell(*cellInfo)

            if action=='copy':
                cellInfo[0].copyCell(cellInfo[1], cellInfo[2],
                                     self.cellInfo[0], self.cellInfo[1],
                                     self.cellInfo[2])

            if action=='create_analogy':
                p1Info = cellInfo[0].getPipelineInfo(cellInfo[1], cellInfo[2])
                p2Info = self.cellInfo[0].getPipelineInfo(self.cellInfo[1],
                                                          self.cellInfo[2])
                if p1Info!=None and p2Info!=None:
                    analogy = SpreadsheetAnalogy()
                    analogy.createAnalogy(p1Info, p2Info)

            if action=='apply_analogy':
                p1Info = cellInfo[0].getPipelineInfo(cellInfo[1], cellInfo[2])
                analogy = SpreadsheetAnalogy()
                newPipeline = analogy.applyAnalogy(p1Info)
                if newPipeline:
                    self.cellInfo[0].executePipelineToCell(newPipeline,
                                                           self.cellInfo[1],
                                                           self.cellInfo[2],
                                                           'Apply Analogy')

        else:
            event.ignore()


    def highlight(self, on=True):
        """ highlight(on: bool) -> None
        Highlight the cell as if being selected

        """
        if on:
            self.innerRubberBand.setGeometry(self.rect())
            self.innerRubberBand.show()
        else:
            self.innerRubberBand.hide()

    def updateVersion(self):
        """ updateVersion() -> None
        Use the performed action of this cell to add back a new
        version to the version tree

        """
        from .spreadsheet_controller import spreadsheetController
        builderWindow = spreadsheetController.getBuilderWindow()
        if builderWindow:
            info = self.cellInfo[0].getCellPipelineInfo(self.cellInfo[1],
                                                        self.cellInfo[2])
            if info:
                info = info[0]
                view = builderWindow.ensureVistrail(info['locator'])
                if view:
                    controller = view.controller
                    controller.change_selected_version(info['version'])
                    controller.perform_param_changes(info['actions'])
                    # controller.performBulkActions(info['actions'])

    def locateVersion(self):
        """ locateVersion() -> None
        Select the version node on the version that has generated this cell

        """
        from .spreadsheet_controller import spreadsheetController
        builderWindow = spreadsheetController.getBuilderWindow()
        if builderWindow:
            info = self.cellInfo[0].getCellPipelineInfo(self.cellInfo[1],
                                                        self.cellInfo[2])
            if info:
                info = info[0]
                view = builderWindow.ensureVistrail(info['locator'])
                if view:
                    view.version_selected(info['version'], True)
                    view.version_view.select_current_version()
                    builderWindow.view_changed(view)
                    w = view.window()
                    # this has no effect
                    w.qactions['history'].trigger()
                    # so we need to use this one
                    view.history_selected()
                    view.activateWindow()


class QCellDragLabel(QtGui.QLabel):
    """
    QCellDragLabel is a pixmap label allowing users to drag it to
    another cell manipulator

    """
    def __init__(self, pixmap, parent=None):
        """ QCellDragLabel(pixmap: QPixmap, parent: QWidget) -> QCellDragLabel
        Construct the pixmap label

        """
        QtGui.QLabel.__init__(self, parent)
        self.setMargin(0)
        self.setSizePolicy(QtGui.QSizePolicy.Maximum, QtGui.QSizePolicy.Maximum)
        self.setPixmap(pixmap)
        self.setScaledContents(True)
        self.setFixedSize(64, 64)
        self.cursorPixmap = pixmap.scaled(self.size())

        self.startPos = None
        self.cellInfo = (None, -1, -1)
        self.action = None

    def updateCellInfo(self, cellInfo):
        """ updateCellInfo(cellInfo: tuple) -> None
        Update cellInfo for mime data while dragging

        """
        self.cellInfo = cellInfo

    def mousePressEvent(self, event):
        """ mousePressEvent(event: QMouseEvent) -> None
        Store the start position for drag event

        """
        self.startPos = QtCore.QPoint(event.pos())
        QtGui.QLabel.mousePressEvent(self, event)

    def mouseMoveEvent(self, event):
        """ mouseMoveEvent(event: QMouseEvent) -> None
        Prepare to drag

        """
        p = event.pos() - self.startPos
        if p.manhattanLength()>=QtGui.QApplication.startDragDistance():
            drag = QtGui.QDrag(self)
            data = QtCore.QMimeData()
            data.cellInfo = self.cellInfo
            data.action = self.action
            data.manipulator = self.parent()
            drag.setMimeData(data)
            drag.setHotSpot(self.cursorPixmap.rect().center())
            drag.setPixmap(self.cursorPixmap)
            drag.start(QtCore.Qt.MoveAction)



#!/usr/bin/env python

# Demonstrate how to use the vtkBoxWidget to control volume rendering
# within the interior of the widget.

import vtk
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()

# Load a volume, use the widget to control what's volume
# rendered. Basically the idea is that the vtkBoxWidget provides a box
# which clips the volume rendering.
v16 = vtk.vtkVolume16Reader()
v16.SetDataDimensions(64, 64)
v16.GetOutput().SetOrigin(0.0, 0.0, 0.0)
v16.SetDataByteOrderToLittleEndian()
v16.SetFilePrefix(VTK_DATA_ROOT+ "/Data/headsq/quarter")
v16.SetImageRange(1, 93)
v16.SetDataSpacing(3.2, 3.2, 1.5)

tfun = vtk.vtkPiecewiseFunction()
tfun.AddPoint(70.0, 0.0)
tfun.AddPoint(599.0, 0)
tfun.AddPoint(600.0, 0)
tfun.AddPoint(1195.0, 0)
tfun.AddPoint(1200, .2)
tfun.AddPoint(1300, .3)
tfun.AddPoint(2000, .3)
tfun.AddPoint(4095.0, 1.0)

ctfun = vtk.vtkColorTransferFunction()
ctfun.AddRGBPoint(0.0, 0.5, 0.0, 0.0)
ctfun.AddRGBPoint(600.0, 1.0, 0.5, 0.5)
ctfun.AddRGBPoint(1280.0, 0.9, 0.2, 0.3)
ctfun.AddRGBPoint(1960.0, 0.81, 0.27, 0.1)
ctfun.AddRGBPoint(4095.0, 0.5, 0.5, 0.5)

compositeFunction = vtk.vtkVolumeRayCastCompositeFunction()

volumeMapper = vtk.vtkVolumeRayCastMapper()
volumeMapper.SetInputConnection(v16.GetOutputPort())
volumeMapper.SetVolumeRayCastFunction(compositeFunction)

volumeProperty = vtk.vtkVolumeProperty()
volumeProperty.SetColor(ctfun)
volumeProperty.SetScalarOpacity(tfun)
volumeProperty.SetInterpolationTypeToLinear()
volumeProperty.ShadeOn()

newvol = vtk.vtkVolume()
newvol.SetMapper(volumeMapper)
newvol.SetProperty(volumeProperty)

outline = vtk.vtkOutlineFilter()
outline.SetInputConnection(v16.GetOutputPort())
outlineMapper = vtk.vtkPolyDataMapper()
outlineMapper.SetInputConnection(outline.GetOutputPort())
outlineActor = vtk.vtkActor()
outlineActor.SetMapper(outlineMapper)

# Create the RenderWindow, Renderer and both Actors
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# The SetInteractor method is how 3D widgets are associated with the
# render window interactor. Internally, SetInteractor sets up a bunch
# of callbacks using the Command/Observer mechanism (AddObserver()).
boxWidget = vtk.vtkBoxWidget()
boxWidget.SetInteractor(iren)
boxWidget.SetPlaceFactor(1.0)

# Add the actors to the renderer, set the background and size
ren.AddActor(outlineActor)
ren.AddVolume(newvol)

ren.SetBackground(0, 0, 0)
renWin.SetSize(300, 300)

# When interaction starts, the requested frame rate is increased.
def StartInteraction(obj, event):
    global renWin
    renWin.SetDesiredUpdateRate(10)

# When interaction ends, the requested frame rate is decreased to
# normal levels. This causes a full resolution render to occur.
def EndInteraction(obj, event):
    global renWin
    renWin.SetDesiredUpdateRate(0.001)

# The implicit function vtkPlanes is used in conjunction with the
# volume ray cast mapper to limit which portion of the volume is
# volume rendered.
planes = vtk.vtkPlanes()
def ClipVolumeRender(obj, event):
    global planes, volumeMapper
    obj.GetPlanes(planes)
    volumeMapper.SetClippingPlanes(planes)
 

# Place the interactor initially. The output of the reader is used to
# place the box widget.
boxWidget.SetInput(v16.GetOutput())
boxWidget.PlaceWidget()
boxWidget.InsideOutOn()
boxWidget.AddObserver("StartInteractionEvent", StartInteraction)
boxWidget.AddObserver("InteractionEvent", ClipVolumeRender)
boxWidget.AddObserver("EndInteractionEvent", EndInteraction)

outlineProperty = boxWidget.GetOutlineProperty()
outlineProperty.SetRepresentationToWireframe()
outlineProperty.SetAmbient(1.0)
outlineProperty.SetAmbientColor(1, 1, 1)
outlineProperty.SetLineWidth(3)

selectedOutlineProperty = boxWidget.GetSelectedOutlineProperty()
selectedOutlineProperty.SetRepresentationToWireframe()
selectedOutlineProperty.SetAmbient(1.0)
selectedOutlineProperty.SetAmbientColor(1, 0, 0)
selectedOutlineProperty.SetLineWidth(3)


iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""Describes the start-up process of VisTrails"""
from __future__ import division

from vistrails.core import debug
from vistrails.core import system
from vistrails.core.configuration import ConfigurationObject
from vistrails.db.domain import DBStartup, DBStartupPackage, \
    DBEnabledPackages, DBDisabledPackages
import vistrails.db.services.io
import vistrails.core.db.io
from vistrails.core.system import get_elementtree_library, \
    get_vistrails_directory, systemType
from vistrails.core.utils import version_string_to_list

import atexit
import copy
import os.path
import re
import shutil
import string
import sys
import tempfile
import vistrails.core.configuration

ElementTree = get_elementtree_library()

################################################################################

class StartupPackage(DBStartupPackage):
    def __init__(self, *args, **kwargs):
        if 'prefix' in kwargs:
            self.prefix = kwargs['prefix']
            del kwargs['prefix']
        else:
            self.prefix = None
        DBStartupPackage.__init__(self, *args, **kwargs)

    @staticmethod
    def convert(_pkg):
        _pkg.__class__ = StartupPackage
        if _pkg.configuration is not None:
            ConfigurationObject.convert(_pkg.configuration)
        _pkg.prefix = None

    def __copy__(self):
        """ __copy__() -> StartupPackage - Returns a clone of itself """ 
        return StartupPackage.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBStartupPackage.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = StartupPackage
        cp.prefix = self.prefix
        return cp

    configuration = DBStartupPackage.db_configuration
    name = DBStartupPackage.db_name

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        return (self.name == other.name and
                self.configuration == other.configuration)

    def __ne__(self, other):
        return not self.__eq__(other)

class VistrailsStartup(DBStartup):
    """
    VistrailsStartup is the class that initializes VisTrails based on
    a configuration. Both application mode (interactive and
    non-interactive) and import mode of VisTrails use this start up
    process to install packages and run .vistrails file. The purpose
    of this class is to separate the initialization process with Qt
    Application
    
    """

    # !!! IMPORTANT: keep logDirectory first!
    DIRECTORIES = [('logDir', 'logs', False),
                   ('userPackageDir', 'userpackages', True),
                   ('subworkflowsDir', 'subworkflows', False),
                   ('thumbs.cacheDir', 'thumbs', False)]
    DOT_VISTRAILS_PREFIX = "$DOT_VISTRAILS"
    DOT_VISTRAILS_PREFIX_RE = re.compile("%s([%s/\\\\]|$)" % 
                                         (re.escape(DOT_VISTRAILS_PREFIX),
                                          os.path.sep))

    first_run = False

    def __init__(self, options_config, command_line_config, 
                 use_dot_vistrails=True):
        """VistrailsStartup(dot_vistrails: str) -> None

        Setup VisTrails configuration and options. dot_vistrals
        indicates the file we will use to load and save configuration
        options; if it is None, options will not be loaded or saved
        for the session.

        """

        DBStartup.__init__(self)
        self.db_enabled_packages = DBEnabledPackages()
        self.db_disabled_packages = DBDisabledPackages()

        self._dot_vistrails = None

        # self._enabled_packages = {}
        # self._disabled_packages = {}

        self.configuration = vistrails.core.configuration.default()

        if ((command_line_config is not None and
                 command_line_config.check('spawned')) or 
                (options_config is not None and 
                 options_config.check('spawned'))):
            # Here we are in 'spawned' mode, i.e. we are running
            # non-interactively as a slave
            # We are going to create a .vistrails directory as a temporary
            # directory and copy a specific configuration file
            # We don't want to load packages that the user might enabled in
            # this machine's configuration file as it would slow down the
            # startup time, but we'll load any needed package without
            # confirmation
            tmpdir = tempfile.mkdtemp(prefix='vt_spawned_')
            @atexit.register
            def clean_dotvistrails():
                shutil.rmtree(tmpdir, ignore_errors=True)
            command_line_config.dotVistrails = tmpdir
            shutil.copyfile(os.path.join(system.vistrails_root_directory(),
                                         'core', 'resources',
                                         'spawned_startup_xml'),
                            os.path.join(tmpdir, 'startup.xml'))
            command_line_config.enablePackagesSilently = True
            command_line_config.errorLog = False
            command_line_config.singleInstance = False

        if use_dot_vistrails:
            if command_line_config is not None and \
               command_line_config.check('dotVistrails'):
                self._dot_vistrails = command_line_config.get('dotVistrails')
            elif options_config is not None and \
                 options_config.check('dotVistrails'):
                self._dot_vistrails = options_config.get('dotVistrails')
            else:
                self._dot_vistrails = self.configuration.get('dotVistrails')
            self.setup_base_dot_vistrails()

        self.load_and_update_configurations(options_config, 
                                            command_line_config)
        self.update_packages()
        self.setup_dot_vistrails()

    @staticmethod
    def convert(_startup):
        _startup.__class__ = VistrailsStartup
        ConfigurationObject.convert(_startup.configuration)
        for _pkg in _startup.db_enabled_packages.db_packages:
            StartupPackage.convert(_pkg)
        for _pkg in _startup.db_disabled_packages.db_packages:
            StartupPackage.convert(_pkg)

    ########################################################################
    # Properties

    configuration = DBStartup.db_configuration
    def _get_enabled_packages(self):
        return self.db_enabled_packages.db_packages_name_index
    enabled_packages = property(_get_enabled_packages)
    def _get_disabled_packages(self):
        return self.db_disabled_packages.db_packages_name_index
    disabled_packages = property(_get_disabled_packages)
    version = DBStartup.db_version

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        def check_packages(self_pkgs, other_pkgs):
            if len(self_pkgs.db_packages) != len(other_pkgs.db_packages):
                return False
            for (p_name, pkg1) in \
                    self_pkgs.db_packages_name_index.iteritems():
                if p_name not in other_pkgs.db_packages_name_index:
                    return False
                else:
                    pkg2 = other_pkgs.db_packages_name_index[p_name]
                    if pkg1 != pkg2:
                        return False
            return True
        return (self.configuration == other.configuration and
                check_packages(self.db_enabled_packages, 
                               other.db_enabled_packages) and
                check_packages(self.db_disabled_packages,
                               other.db_disabled_packages))

    def get_startup_xml_fname(self):
        return os.path.join(self._dot_vistrails, 'startup.xml')

    def setup_base_dot_vistrails(self):
        self.create_dot_vistrails_if_necessary()
        self.create_startupxml_if_needed()

    def load_and_update_configurations(self, options_config, 
                                       command_line_config):
        # load options from startup.xml
        self.load_persisted_startup()

        # set up temporary configuration with options and command-line
        self.temp_configuration = copy.copy(self.configuration)

        if options_config is not None:
            self.temp_configuration.update(options_config)
        if command_line_config is not None:
            self.temp_configuration.update(command_line_config)

    def setup_init_file(self, dir_name):
        name = os.path.join(dir_name, '__init__.py')
        try:
            f = open(name, 'w')
            f.write('pass\n')
            f.close()
        except:
            msg = ("""Failed to create file '%s'. This could indicate a
            rare combination of a race condition and a permissions problem.
            Please make sure it is writable.""" % name)
            debug.critical(msg)
            sys.exit(1)

    def expand_vt_path(self, dir_name):
        m = self.DOT_VISTRAILS_PREFIX_RE.match(dir_name)
        if m:
            if self._dot_vistrails is None:
                return None
            else:
                # could be "$DOT_VISTRAILS" or "$DOT_VISTRAILS/userpackages"
                prefix_len = len(self.DOT_VISTRAILS_PREFIX) + len(m.groups()[0])
                if prefix_len < len(dir_name):
                    return os.path.join(self._dot_vistrails, 
                                        dir_name[prefix_len:])
                else:
                    return self._dot_vistrails
        return dir_name

    def create_directory(self, dir_name):
        if not os.path.isdir(dir_name):
            debug.warning('Will try to create directory "%s"' % dir_name)
            try:
                os.mkdir(dir_name)
                return True
            except Exception, e:
                msg = ("Failed to create directory: '%s'."
                       "This could be an indication of a permissions problem."
                       "Make sure directory '%s' in writable." %
                       (str(e), dir_name))
                debug.critical(msg)
                sys.exit(1)
        return False

    def setup_directory(self, config_key, default_dir, setup_init_file=False):
        abs_dir_name = get_vistrails_directory(config_key, 
                                               self.temp_configuration)        
        if abs_dir_name is not None:
            self.create_directory(abs_dir_name)
        if abs_dir_name is not None and setup_init_file:
            self.setup_init_file(abs_dir_name)
        return abs_dir_name

    def setup_log_file(self, abs_dir_name):
        import vistrails.core.system
        version = vistrails.core.system.vistrails_version()
        version = version.replace(".", "_")

        if self.temp_configuration.errorLog:
            log_fname = os.path.join(abs_dir_name, 'vistrails_%s.log' % version)
            if log_fname is not None:
                debug.DebugPrint.getInstance().set_logfile(log_fname)

    def setup_debug(self):
        if (self.temp_configuration.has('debugLevel') and
            self.temp_configuration.debugLevel != -1):
            verbose = self.temp_configuration.debugLevel
            if verbose < 0:
                msg = ("""Don't know how to set verboseness level to %s - "
                       "setting to the lowest one I know of: 0""" % verbose)
                debug.critical(msg)
                verbose = self.temp_configuration.debugLevel = 0
            if verbose > 2:
                msg = ("""Don't know how to set verboseness level to %s - "
                       "setting to the highest one I know of: 2""" % verbose)
                debug.critical(msg)
                verbose = self.temp_configuration.debugLevel = 2
            dbg = debug.DebugPrint.getInstance()
            levels = [dbg.WARNING, dbg.INFO, dbg.DEBUG]
            dbg.set_message_level(levels[verbose])
            debug.log("Set verboseness level to %s" % verbose)


    def setup_dot_vistrails(self):
        for i, args in enumerate(self.DIRECTORIES):
            abs_dir_name = self.setup_directory(*args)
            # Brittle, know that log directory is first
            if i == 0:
                self.setup_log_file(abs_dir_name)
        self.setup_debug()

    def create_dot_vistrails_if_necessary(self):
        if os.path.exists(self._dot_vistrails):
            if not os.path.isdir(self._dot_vistrails):
                raise ValueError('The .vistrails directory cannot be used or '
                                 'created because the specified path "%s" is '
                                 'a file not a directory.' % 
                                 self._dot_vistrails)
            else:
                return

        debug.log('Will try to create default directory')
        try:
            os.mkdir(self._dot_vistrails)
            debug.log('Succeeded!')
        except:
            debug.critical("""Failed to create initialization directory.
                    This could be an indication of a permissions problem.
                    Make sure parent directory of '%s' is writable."""
                    % self._dot_vistrails)
            raise

    def update_packages(self):
        # make sure basic_modules and abstraction are enabled
        self.set_package_to_enabled('basic_modules')
        self.set_package_to_enabled('abstraction')
        self.enabled_packages['basic_modules'].prefix = \
                                            "vistrails.core.modules."
        self.enabled_packages['abstraction'].prefix = \
                                            "vistrails.core.modules."

    def load_persisted_startup(self):
        if self._dot_vistrails is not None:
            fname = self.get_startup_xml_fname()
            other = vistrails.core.db.io.load_startup(fname)
            self.configuration.update(other.configuration)
            self.db_enabled_packages = copy.copy(other.db_enabled_packages)
            self.db_disabled_packages = copy.copy(other.db_disabled_packages)

    def save_persisted_startup(self):
        if self._dot_vistrails is not None:
            fname = self.get_startup_xml_fname()
            vistrails.core.db.io.save_startup(self, fname)

    def get_pkg_startup_info(self, codepath):
        if codepath in self.enabled_packages:
            return self.enabled_packages[codepath]
        elif codepath in self.disabled_packages:
            return self.disabled_packages[codepath]
        return None

    def get_pkg_configuration(self, codepath):
        startup_info = self.get_pkg_startup_info(codepath)
        if startup_info is not None:
            return startup_info.configuration
        return None

    def persist_pkg_configuration(self, codepath, config):
        startup_info = self.get_pkg_startup_info(codepath)
        if startup_info is not None:
            startup_info.configuration = config
        else:
            startup_info = StartupPackage(name=codepath,
                                          configuration=config)
            self.db_disabled_packages.db_add_package(startup_info)
        self.save_persisted_startup()

    def set_package_enabled(self, codepath, enabled=True):
        package = None
        was_enabled = False
        was_disabled = False
        if codepath in self.enabled_packages:
            package = self.enabled_packages[codepath]
            was_enabled = True
        if codepath in self.disabled_packages:
            if package is None:
                package = self.disabled_packages[codepath]
            was_disabled = True
        if package is None:
            package = StartupPackage(name=codepath)

        def update_package_list(update_obj):
            # this is a kludge since the db code doesn't support non-keyed
            # deletion and name isn't a key right now
            # FIXME may cause issues with relational storage
            old_packages = update_obj.db_packages
            update_obj.db_packages = []
            update_obj.db_packages_name_index = {}
            for p in old_packages:
                if p.name != codepath:
                    update_obj.db_add_package(p)
            
        if was_enabled:
            update_package_list(self.db_enabled_packages)
        if was_disabled:
            update_package_list(self.db_disabled_packages)

        if enabled:
            self.db_enabled_packages.db_add_package(package)
        else:
            self.db_disabled_packages.db_add_package(package)

    def set_package_to_enabled(self, codepath):
        self.set_package_enabled(codepath, True)

    def set_package_to_disabled(self, codepath):
        self.set_package_enabled(codepath, False)

    def create_startupxml_if_needed(self):
        needs_create = True
        fname = self.get_startup_xml_fname()
        if os.path.isfile(fname):
            try:
                tree = ElementTree.parse(fname)
                startup_version = \
                    vistrails.db.services.io.get_version_for_xml(tree.getroot())
                version_list = version_string_to_list(startup_version)
                if version_list >= [0,1]:
                    needs_create = False
            except:
                debug.warning("Unable to read startup.xml file, "
                              "creating a new one")

        if needs_create:
            root_dir = system.vistrails_root_directory()
            origin = os.path.join(root_dir, 'core','resources',
                                  'default_vistrails_startup_xml')
            try:
                shutil.copyfile(origin, fname)
                debug.log('Succeeded!')
                self.first_run = True
            except:
                debug.critical("""Failed to copy default configuration
                file to %s. This could be an indication of a
                permissions problem. Please make sure '%s' is writable."""
                               % (fname, self._dot_vistrails))
                raise

import unittest
import stat

class TestStartup(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Mock DebugPrint#set_logfile()
        def set_logfile(self, filename):
            cls._log_filename = filename
        cls._old_debugprint_set_logfile = debug.DebugPrint.set_logfile
        debug.DebugPrint.set_logfile = set_logfile

    @classmethod
    def tearDownClass(cls):
        # Un-mock DebugPrint#set_logfile()
        debug.DebugPrint.set_logfile = cls._old_debugprint_set_logfile

    def check_structure(self, dir_name):
        self.assertTrue(os.path.isdir(dir_name))
        self.assertTrue(os.path.isfile(os.path.join(dir_name, 
                                                    'startup.xml')))
        for t in VistrailsStartup.DIRECTORIES:
            a_dir = os.path.join(dir_name, t[1])
            self.assertTrue(os.path.isdir(a_dir),
                            'Directory "%s" does not exist' % a_dir)
            if t[2]:
                self.assertTrue(os.path.isfile(os.path.join(a_dir, 
                                                            '__init__.py')))

        import vistrails.core.system
        version = vistrails.core.system.vistrails_version()
        version = version.replace(".", "_")
        self.assertEqual(
                self._log_filename,
                os.path.join(dir_name, 'logs', 'vistrails_%s.log' % version))

    def test_simple_create(self):
        dir_name = tempfile.mkdtemp()
        options_config = ConfigurationObject(dotVistrails=dir_name)
        try:
            startup = VistrailsStartup(options_config, None)
            self.check_structure(dir_name)
        finally:
            shutil.rmtree(dir_name)

    def test_create_dir_create(self):
        outer_dir_name = tempfile.mkdtemp()
        dir_name = os.path.join(outer_dir_name, '.vistrails')
        cl_config = ConfigurationObject(dotVistrails=dir_name)
        try:
            startup = VistrailsStartup(None, cl_config)
            self.check_structure(dir_name)
        finally:
            shutil.rmtree(outer_dir_name)
        
    def test_config_override(self):
        dir_name = tempfile.mkdtemp()
        user_pkg_dir = tempfile.mkdtemp()
        abstractions_dir = tempfile.mkdtemp()
        thumbs_dir = tempfile.mkdtemp()
        log_dir = tempfile.mkdtemp()
        config = vistrails.core.configuration.default()
        config.dotVistrails = dir_name
        config.userPackageDir = user_pkg_dir
        config.subworkflowsDir= abstractions_dir
        config.thumbs.cacheDir = thumbs_dir
        config.logDir = log_dir
        try:
            startup = VistrailsStartup(config, None)
            self.assertTrue(os.path.isfile(os.path.join(dir_name, 
                                                        'startup.xml')))
            for t in VistrailsStartup.DIRECTORIES:
                a_dir = os.path.join(dir_name, t[1])
                self.assertFalse(os.path.isdir(a_dir),
                                'Directory "%s" exists' % a_dir)
        finally:
            shutil.rmtree(dir_name)
            shutil.rmtree(user_pkg_dir)
            shutil.rmtree(abstractions_dir)
            shutil.rmtree(thumbs_dir)
            shutil.rmtree(log_dir)

    def test_config_override_create(self):
        dir_name = tempfile.mkdtemp()
        outer_user_pkg_dir = tempfile.mkdtemp()
        user_pkg_dir = os.path.join(outer_user_pkg_dir, 'userpackages')
        outer_abstractions_dir = tempfile.mkdtemp()
        abstractions_dir = os.path.join(outer_abstractions_dir, 'subworkflows')
        outer_thumbs_dir = tempfile.mkdtemp()
        thumbs_dir = os.path.join(outer_thumbs_dir, 'thumbs')
        outer_log_dir = tempfile.mkdtemp()
        log_dir = os.path.join(outer_log_dir, 'logs')
        config = vistrails.core.configuration.default()
        config.dotVistrails = dir_name
        config.userPackageDir = user_pkg_dir
        config.subworkflowsDir = abstractions_dir
        config.thumbs.cacheDir = thumbs_dir
        config.logDir = log_dir
        try:
            startup = VistrailsStartup(config, None)
            self.assertTrue(os.path.isfile(os.path.join(dir_name, 
                                                        'startup.xml')))
            for t in VistrailsStartup.DIRECTORIES:
                a_dir = os.path.join(dir_name, t[1])
                self.assertFalse(os.path.isdir(a_dir),
                                'Directory "%s" exists' % a_dir)
            self.assertTrue(os.path.isdir(user_pkg_dir))
            self.assertTrue(os.path.isdir(abstractions_dir))
            self.assertTrue(os.path.isdir(thumbs_dir))
            self.assertTrue(os.path.isdir(log_dir))
        finally:
            shutil.rmtree(dir_name)
            shutil.rmtree(outer_user_pkg_dir)
            shutil.rmtree(outer_abstractions_dir)
            shutil.rmtree(outer_thumbs_dir)
            shutil.rmtree(outer_log_dir)
    
    def test_default_startup_xml(self):
        dir_name = tempfile.mkdtemp()
        config = ConfigurationObject(dotVistrails=dir_name)
        try:
            startup = VistrailsStartup(config, None)
            self.assertTrue(startup.configuration.executionLog)
            self.assertTrue(startup.configuration.autoSave)
            self.assertTrue(startup.temp_configuration.autoSave)
        finally:
            shutil.rmtree(dir_name)

    def test_cannot_create(self):
        (fd, fname) = tempfile.mkstemp()
        os.close(fd)
        config = ConfigurationObject(dotVistrails=fname)
        try:
            with self.assertRaises(ValueError):
                startup = VistrailsStartup(config, None)
        finally:
            os.unlink(fname)
        
    def test_permissions(self):
        if systemType in ['Windows', 'Microsoft']:
            self.skipTest("chmod on Windows is limited")
        dir_name = tempfile.mkdtemp()
        config = ConfigurationObject(dotVistrails=dir_name)
        try:
            os.chmod(dir_name, stat.S_IRUSR)
            with self.assertRaises(IOError):
                startup = VistrailsStartup(config, None)
        finally:
            os.chmod(dir_name, stat.S_IRWXU)
            shutil.rmtree(dir_name)

    def test_load_old_startup_xml(self):
        # has old nested shell settings that don't match current naming
        startup_tmpl = os.path.join(system.vistrails_root_directory(),
                                    'tests', 'resources',
                                    'startup-0.1.xml.tmpl')
        f = open(startup_tmpl, 'r')
        template = string.Template(f.read())
        
        startup_dir = tempfile.mkdtemp(prefix="vt_startup")
        old_startup_fname = os.path.join(startup_dir, "startup.xml")
        with open(old_startup_fname, 'w') as f:
            f.write(template.substitute({'startup_dir': startup_dir}))

        startup1 = vistrails.core.db.io.load_startup(old_startup_fname)

        (h, fname) = tempfile.mkstemp(suffix=".xml")
        os.close(h)
        try:
            vistrails.core.db.io.save_startup(startup1, fname)
            startup2 = vistrails.core.db.io.load_startup(fname)
            self.assertEqual(startup1, startup2)
        finally:
            os.remove(fname)
            shutil.rmtree(startup_dir)

    # def test_load_packages(self):
    #     from vistrails.core.system import default_dot_vistrails
    #     dir_name = default_dot_vistrails()
    #     config = ConfigurationObject(dotVistrails=dir_name)
    #     startup = VistrailsStartup(config, None)
        
        
if __name__ == '__main__':
    unittest.main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from xml.auto_gen import XMLDAOListBase
from sql.auto_gen import SQLDAOListBase
from vistrails.core.system import get_elementtree_library

from vistrails.db import VistrailsDBException
from vistrails.db.versions.v0_9_1 import version as my_version

ElementTree = get_elementtree_library()


class DAOList(dict):
    def __init__(self):
        self['xml'] = XMLDAOListBase()
        self['sql'] = SQLDAOListBase()

    def parse_xml_file(self, filename):
        return ElementTree.parse(filename)

    def write_xml_file(self, filename, tree):
        tree.write(filename)

    def read_xml_object(self, vtType, node):
        return self['xml'][vtType].fromXML(node)

    def write_xml_object(self, obj, node=None):
        res_node = self['xml'][obj.vtType].toXML(obj, node)
        return res_node
        
    def open_from_xml(self, filename, vtType, tree=None):
        """open_from_xml(filename) -> DBVistrail"""
        if tree is None:
            tree = self.parse_xml_file(filename)
        vistrail = self.read_xml_object(vtType, tree.getroot())
        return vistrail

    def save_to_xml(self, obj, filename, tags, version=None):
        """save_to_xml(obj : object, filename: str, tags: dict,
                       version: str) -> None
    
        """
        root = self.write_xml_object(obj)
        if version is None:
            version = my_version
        root.set('version', version)
        for k, v in tags.iteritems():
            root.set(k, v)
        tree = ElementTree.ElementTree(root)
        self.write_xml_file(filename, tree)

    def serialize(self, object):
        root = self.write_xml_object(object)
        return ElementTree.tostring(root)

    def unserialize(self, str, obj_type):
        try:
            root = ElementTree.fromstring(str)
            return self.read_xml_object(obj_type, root)
        except SyntaxError, e:
            msg = "Invalid VisTrails serialized object %s" % str
            raise VistrailsDBException(msg)
            return None



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core import debug
from vistrails.core.modules.basic_modules import create_constant, get_module
from vistrails.core.modules.vistrails_module import Module, ModuleError, \
    ModuleConnector, InvalidOutput, ModuleSuspended, ModuleWasSuspended
from vistrails.core.modules.basic_modules import Boolean, String, Integer, \
    Float, Constant, List
from vistrails.core.modules.module_registry import get_module_registry
from vistrails.core.vistrail.port_spec import PortSpec

import copy
from itertools import izip

###############################################################################
## Fold Operator

class Fold(Module):
    """Fold is the base class for List-reducing modules.

    It can be used to easily implement a module that takes a List and
    aggregates its element one by one to get the final result, such as Sum.

    To use it, create a subclass and override the setInitialValue() and
    operation() methods.
    """

    def __init__(self):
        Module.__init__(self)

    def compute(self):
        """The compute method for the Fold."""

        self.setInitialValue()
        self.partialResult = self.initialValue
        self.elementResult = None

        for element in self.get_input('InputList'):
            self.element = element
            self.operation()

        self.set_output('Result', self.partialResult)

    def setInitialValue(self): # pragma: no cover
        """This method defines the initial value of the Fold structure. It must
        be defined before the operation() method."""

        pass

    def operation(self): # pragma: no cover
        """This method defines the interaction between the current element of
        the list and the previous iterations' result."""

        pass

###############################################################################

class FoldWithModule(Fold):
    """Implementation of Fold that uses another module as its operation.

    This can be used to create structures like Map or Filter, where another
    module will be called with each element of the list to retrieve something
    that this module will use.
    """

    def update_upstream(self):
        """A modified version of the update_upstream method."""

        # everything is the same except that we don't update the module on
        # FunctionPort
        suspended = []
        was_suspended = None
        for port_name, connector_list in self.inputPorts.iteritems():
            if port_name == 'FunctionPort':
                for connector in connector_list:
                    try:
                        connector.obj.update_upstream()
                    except ModuleWasSuspended, e:
                        was_suspended = e
                    except ModuleSuspended, e:
                        suspended.append(e)
            else:
                for connector in connector_list:
                    try:
                        connector.obj.update()
                    except ModuleWasSuspended, e:
                        was_suspended = e
                    except ModuleSuspended, e:
                        suspended.append(e)
        if len(suspended) == 1:
            raise suspended[0]
        elif suspended:
            raise ModuleSuspended(
                    self,
                    "multiple suspended upstream modules",
                    children=suspended)
        elif was_suspended is not None:
            raise was_suspended
        for port_name, connectorList in list(self.inputPorts.items()):
            if port_name != 'FunctionPort':
                for connector in connectorList:
                    if connector.obj.get_output(connector.port) is \
                            InvalidOutput: # pragma: no cover
                        self.remove_input_connector(port_name, connector)

    def updateFunctionPort(self):
        """
        Function to be used inside the updateUsptream method of the
        FoldWithModule module. It updates the modules connected to the
        FunctionPort port.
        """
        nameInput = self.get_input('InputPort')
        nameOutput = self.get_input('OutputPort')
        rawInputList = self.get_input('InputList')

        # Create inputList to always have iterable elements
        # to simplify code
        if len(nameInput) == 1:
            element_is_iter = False
            inputList = [[element] for element in rawInputList]
        else:
            element_is_iter = True
            inputList = rawInputList
        suspended = []
        loop = self.logging.begin_loop_execution(self, len(inputList))
        ## Update everything for each value inside the list
        for i, element in enumerate(inputList):
            self.logging.update_progress(self, float(i)/len(inputList))
            if element_is_iter:
                self.element = element
            else:
                self.element = element[0]
            do_operation = True
            for connector in self.inputPorts.get('FunctionPort'):
                module = copy.copy(connector.obj)

                if not self.upToDate: # pragma: no branch
                    ## Type checking
                    if i == 0:
                        self.typeChecking(module, nameInput, inputList)

                    module.upToDate = False
                    module.computed = False

                    self.setInputValues(module, nameInput, element, i)

                loop.begin_iteration(module, i)

                try:
                    module.update()
                except ModuleSuspended, e:
                    suspended.append(e)
                    do_operation = False
                    loop.end_iteration(module)
                    continue

                loop.end_iteration(module)

                ## Getting the result from the output port
                if nameOutput not in module.outputPorts:
                    raise ModuleError(module,
                                      'Invalid output port: %s' % nameOutput)
                self.elementResult = module.get_output(nameOutput)
            if do_operation:
                self.operation()

            self.logging.update_progress(self, i * 1.0 / len(inputList))

        if suspended:
            raise ModuleSuspended(
                    self,
                    "function module suspended in %d/%d iterations" % (
                            len(suspended), len(inputList)),
                    children=suspended)
        loop.end_loop_execution()

    def compute(self):
        """The compute method for the Fold."""

        self.setInitialValue()
        self.partialResult = self.initialValue
        self.elementResult = None

        self.updateFunctionPort()

        self.set_output('Result', self.partialResult)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from .gmap_cell import _modules as gmap_modules
from .vis import _modules as vis_modules
from .identifiers import identifier

from vistrails.core.modules.utils import create_port_spec_string
from vistrails.core.vistrail.port import Port
from vistrails.core.vistrail.port_spec import PortSpec
from vistrails.core.vistrail.port_spec_item import PortSpecItem

_modules = gmap_modules + vis_modules

def handle_module_upgrade_request(controller, module_id, pipeline):
    from vistrails.core.modules.module_descriptor import ModuleDescriptor
    from vistrails.core.upgradeworkflow import UpgradeWorkflowHandler, \
        UpgradePackageRemap, UpgradeModuleRemap
    from vistrails.core.modules.basic_modules import identifier as basic_pkg

    def remap_functions(old_module, new_module, function_remap):
        # FIXME need use_registry flag passed through!
        use_registry = True
        function_ops = []
        for function in old_module.functions:
            if function.name in function_remap:
                remap = function_remap[function.name]
                if remap is None:
                    # don't add the function back in
                    continue
                elif not isinstance(remap, basestring):
                    function_ops.extend(remap(function, new_module))
                    continue
                else:
                    function_name = remap

                if len(function.parameters) > 0:
                    new_param_vals, aliases = zip(*[(p.strValue, p.alias)
                                                    for p in function.parameters])
                else:
                    new_param_vals = []
                    aliases = []
                if use_registry:
                    function_port_spec = function_name
                else:
                    def mk_psi(pos):
                        psi = PortSpecItem(module="Module", package=basic_pkg,
                                           namespace="", pos=pos)
                        return psi
                    n_items = len(new_param_vals)
                    function_port_spec = PortSpec(name=function_name,
                                                  items=[mk_psi(i)
                                                         for i in xrange(n_items)])
                new_function = controller.create_function(new_module,
                                                          function_port_spec,
                                                          new_param_vals,
                                                          aliases)
                new_module.add_function(new_function)

        if None in function_remap:
            # used to add new functions
            remap = function_remap[None]
            function_ops.extend(remap(None, new_module))
        return function_ops

    def remap_dst_connections(old_module, new_module, port_remap):
        # FIXME need use_registry flag passed through!
        use_registry = True
        create_new_connection = UpgradeWorkflowHandler.create_new_connection

        ops = []
        for _, conn_id in pipeline.graph.edges_to(old_module.id):
            old_conn = pipeline.connections[conn_id]
            if old_conn.destination.name in port_remap:
                remap = port_remap[old_conn.destination.name]
                if remap is None:
                    # don't add this connection back in
                    continue
                elif not isinstance(remap, basestring):
                    ops.extend(remap(old_conn, new_module))
                    continue
                else:
                    destination_name = remap

                old_src_module = pipeline.modules[old_conn.source.moduleId]
                if use_registry:
                    destination_port = destination_name
                else:
                    destination_port = Port(name=destination_name,
                                            type='destination',
                                            signature=create_port_spec_string(
                                                [(basic_pkg, 'Variant', '')]))

                new_conn = create_new_connection(controller,
                                                 old_src_module,
                                                 old_conn.source,
                                                 new_module,
                                                 destination_port)
                ops.append(('add', new_conn))
        return ops

    def insert_vis(vis_name, vis_port_remap):
        def remap_vis(old_conn, new_cell_module):
            ops = []
            old_src_module = pipeline.modules[old_conn.source.moduleId]
            old_cell_module = pipeline.modules[old_conn.destination.moduleId]

            new_x = (old_src_module.location.x + new_cell_module.location.x)/2.0
            new_y = (old_src_module.location.y + new_cell_module.location.y)/2.0

            new_vis_desc = ModuleDescriptor(package=identifier,
                                            name=vis_name,
                                            version='0.3.0')
            new_vis_module = \
                controller.create_module_from_descriptor(new_vis_desc,
                                                         new_x, new_y)

            function_ops = remap_functions(old_cell_module,
                                           new_vis_module,
                                           vis_port_remap)
            ops.append(('add', new_vis_module))
            ops.extend(function_ops)
            ops.extend(remap_dst_connections(old_cell_module,
                                             new_vis_module,
                                             vis_port_remap))

            create_new_connection = UpgradeWorkflowHandler.create_new_connection
            new_conn_1 = create_new_connection(controller,
                                               old_src_module,
                                               old_conn.source,
                                               new_vis_module,
                                               "table")
            ops.append(('add', new_conn_1))
            new_conn_2 = create_new_connection(controller,
                                               new_vis_module,
                                               "self",
                                               new_cell_module,
                                               "layers")
            ops.append(('add', new_conn_2))
            return ops

        # returns the actual remap function
        return remap_vis

    # zoom gets moved for free from old cell to new cell
    remap = UpgradePackageRemap()
    def add_legacy(fname, module):
        new_function = controller.create_function(module,
                                                  "allowLegacy",
                                                  ["True"])
        return [('add', new_function, 'module', module.id)]
    remap.add_module_remap(UpgradeModuleRemap('0.1.0', '0.3.0', '0.3.0',
                                              new_module="GMapCell",
                                              module_name="GMapCell",
                            dst_port_remap={'table': insert_vis("GMapSymbols",
                                            {None: add_legacy}),
                                            'colormapName': None}))
    remap.add_module_remap(UpgradeModuleRemap('0.1.0', '0.3.0', '0.3.0',
                                              new_module="GMapCell",
                                              module_name="GMapHeatmapCell",
                            dst_port_remap={'table': insert_vis("GMapHeatmap",
                                            {'dissipating': 'dissipating',
                                             'maxIntensity': 'maxIntensity',
                                             'opacity': 'opacity',
                                             'radius': 'radius'}),
                                            'dissipating': None,
                                            'maxIntensity': None,
                                            'opacity': None,
                                            'radius': None,
                                            }))
    remap.add_module_remap(UpgradeModuleRemap('0.1.0', '0.3.0', '0.3.0',
                                              new_module="GMapCell",
                                              module_name="GMapCircleCell",
                            dst_port_remap={'table': insert_vis("GMapCircles",
                                            {'strokeColor': 'strokeColor',
                                             'strokeWeight': 'strokeWeight',
                                             'strokeOpacity': 'strokeOpacity',
                                             'fillColor': 'fillColor',
                                             'fillOpacity': 'fillOpacity'}),
                                            'strokeColor': None,
                                            'strokeWeight': None,
                                            'strokeOpacity': None,
                                            'fillColor': None,
                                            'fillOpacity': None,
                                            }))
    remap.add_module_remap(UpgradeModuleRemap('0.1.0', '0.3.0', '0.3.0',
                                              new_module="GMapCell",
                                              module_name="GMapSymbolCell",
                            dst_port_remap={'table': insert_vis("GMapSymbols",
                                            {'strokeColor': 'strokeColor',
                                             'strokeWeight': 'strokeWeight',
                                             'strokeOpacity': 'strokeOpacity',
                                             'fillStartColor': 'fillStartColor',
                                             'fillEndColor': 'fillEndColor',
                                             'fillOpacity': 'fillOpacity',
                                             'scale': 'scale'}),
                                            'strokeColor': None,
                                            'strokeWeight': None,
                                            'strokeOpacity': None,
                                            'fillStartColor': None,
                                            'fillEndColor': None,
                                            'fillOpacity': None,
                                            'scale': None,
                                        }))

    return UpgradeWorkflowHandler.remap_module(controller, module_id, pipeline,
                                               remap)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.db.versions.v1_0_2.domain import DBVistrail, DBAnnotation, \
                                      DBWorkflow, DBLog, DBRegistry, \
                                      DBPortSpec, DBAdd, DBChange, DBDelete
from vistrails.core import debug
from vistrails.core.system import get_elementtree_library
ElementTree = get_elementtree_library()

import unittest

id_scope = None

def update_portSpec(old_obj, translate_dict):
    global id_scope
    
    sigs = []
    defaults = []
    labels = []
    for psi in sorted(old_obj.db_portSpecItems, key=lambda x: x.db_pos):
        sigs.append((psi.db_package, psi.db_module, psi.db_namespace))
        defaults.append(psi.db_default)
        labels.append(psi.db_label)
    new_obj = DBPortSpec.update_version(old_obj, translate_dict)
    sigstring = '(' + ','.join('%s:%s%s' %
                               (s[0], s[1], ":%s" % s[2] if s[2] else "")
                               for s in sigs) + ')'
    new_obj.db_sigstring = sigstring
    if all(not d for d in defaults):
        new_obj.db_defaults = None
    else:
        new_obj.db_defaults = unicode(defaults)
    if all(not label for label in labels):
        new_obj.db_labels = None
    else:
        new_obj.db_labels = unicode(labels)
        
    return new_obj

def update_portSpecs(old_obj, translate_dict):
    new_port_specs = []
    for port_spec in old_obj.db_portSpecs:
        new_port_specs.append(update_portSpec(port_spec, translate_dict))
    return new_port_specs

def update_portSpec_op(old_obj, translate_dict):
    return update_portSpec(old_obj.db_data, translate_dict)

def translateVistrail(_vistrail):
    """ Translate new DBVistrailVariable based vistrail variables to old
         annotation based type """
    global id_scope
    
    def update_workflow(old_obj, trans_dict):
        return DBWorkflow.update_version(old_obj.db_workflow, 
                                         trans_dict, DBWorkflow())

    def update_operations(old_obj, trans_dict):
        new_ops = []
        for obj in old_obj.db_operations:
            if obj.vtType == 'delete':
                new_ops.append(DBDelete.update_version(obj, trans_dict))
            elif obj.vtType == 'add':
                if obj.db_what == 'portSpec':
                    trans_dict['DBAdd'] = {'data': update_portSpec_op}
                    new_op = DBAdd.update_version(obj, trans_dict)
                    new_ops.append(new_op)
                    del trans_dict['DBAdd']
                else:
                    new_op = DBAdd.update_version(obj, trans_dict)
                    new_ops.append(new_op)
            elif obj.vtType == 'change':
                if obj.db_what == 'portSpec':
                    trans_dict['DBChange'] = {'data': update_portSpec_op}
                    new_op = DBChange.update_version(obj, trans_dict)
                    new_ops.append(new_op)
                    del trans_dict['DBChange']
                else:
                    new_op = DBChange.update_version(obj, trans_dict)
                    new_ops.append(new_op)
        return new_ops

    vistrail = DBVistrail()
    id_scope = vistrail.idScope

    def update_annotations(old_obj, trans_dict):
        new_annotations = []
        for a in old_obj.db_annotations:
            new_annotations.append(DBAnnotation.update_version(a, 
                                                               translate_dict))
            id_scope.updateBeginId(DBAnnotation.vtType, a.db_id)
        vars = {}
        for var in old_obj.db_vistrailVariables:
            descriptor = (var.db_package, var.db_module, var.db_namespace)
            vars[var.db_name] = (var.db_uuid, descriptor, var.db_value)
        if vars:
            new_id = id_scope.getNewId(DBAnnotation.vtType)
            annotation = DBAnnotation(id=new_id, key='__vistrail_vars__', 
                                      value=str(vars))
            new_annotations.append(annotation)

        return new_annotations

    translate_dict = {'DBModule': {'portSpecs': update_portSpecs},
                      'DBModuleDescriptor': {'portSpecs': update_portSpecs},
                      'DBAction': {'operations': update_operations},
                      'DBGroup': {'workflow': update_workflow},
                      'DBVistrail': {'annotations': update_annotations},
                      }

    vistrail = DBVistrail.update_version(_vistrail, translate_dict, vistrail)

    if _vistrail.db_parameter_explorations:
        debug.warning(("Vistrail contains %s parameter explorations that "
                      "cannot be converted") % len(_vistrail.db_parameter_explorations))

    vistrail.db_version = '1.0.2'
    return vistrail

def translateWorkflow(_workflow):
    def update_workflow(old_obj, translate_dict):
        return DBWorkflow.update_version(old_obj.db_workflow, translate_dict)
    translate_dict = {'DBModule': {'portSpecs': update_portSpecs},
                      'DBGroup': {'workflow': update_workflow}}
    workflow = DBWorkflow.update_version(_workflow, translate_dict)

    workflow.db_version = '1.0.2'
    return workflow

def translateLog(_log):
    translate_dict = {}
    log = DBLog.update_version(_log, translate_dict)
    log.db_version = '1.0.2'
    return log

def translateRegistry(_registry):
    global id_scope
    translate_dict = {'DBModuleDescriptor': {'portSpecs': update_portSpecs}}
    registry = DBRegistry()
    id_scope = registry.idScope
    vistrail = DBRegistry.update_version(_registry, translate_dict, registry)
    registry.db_version = '1.0.2'
    return registry

class TestTranslate(unittest.TestCase):
    def testParamexp(self):
        """test translating parameter explorations from 1.0.3 to 1.0.2"""
        from vistrails.db.services.io import open_bundle_from_zip_xml
        from vistrails.core.system import vistrails_root_directory
        import os
        (save_bundle, vt_save_dir) = open_bundle_from_zip_xml(DBVistrail.vtType, \
                        os.path.join(vistrails_root_directory(),
                        'tests/resources/paramexp-1.0.3.vt'))
        vistrail = translateVistrail(save_bundle.vistrail)
        # paramexps cannot be downgraded but should produce a warning
        
    def testVistrailvars(self):
        """test translating vistrail variables from 1.0.3 to 1.0.2"""
        from vistrails.db.services.io import open_bundle_from_zip_xml
        from vistrails.core.system import vistrails_root_directory
        import os
        (save_bundle, vt_save_dir) = open_bundle_from_zip_xml(DBVistrail.vtType, \
                        os.path.join(vistrails_root_directory(),
                        'tests/resources/visvar-1.0.3.vt'))
        vistrail = translateVistrail(save_bundle.vistrail)
        visvars = vistrail.db_annotations_key_index['__vistrail_vars__']
        self.assertTrue(visvars.db_value)

if __name__ == '__main__':
    from vistrails.gui.application import start_application
    v = start_application({'batch': True,
                           'nologger': True,
                           'singleInstance': False,
                           'fixedSpreadsheetCells': True})
    unittest.main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This file contains a dialog and widgets related to the module documentation
dialog, which displays the available documentation for a given VisTrails module.

QMethodDocumentation
"""
from __future__ import division

from PyQt4 import QtCore, QtGui
from vistrails.core.vistrail.port import PortEndPoint
from vistrails.core.utils import VistrailsInternalError

class QPortDocumentation(QtGui.QDialog):
    """
    QPortDocumentation is a dialog for showing port documentation. duh.

    """
    def __init__(self, module, port_type, port_name, parent=None):
        QtGui.QDialog.__init__(self, parent)

        if not module.has_port_spec(port_name, port_type):
            doc = None
        else:
            port_spec = module.get_port_spec(port_name, port_type)
            doc = port_spec.docstring()
            if doc is None:
                descriptor = module.module_descriptor
                # try the old method of accessing documentation
                if port_type == 'output':
                    call_ = descriptor.module.provide_output_port_documentation
                elif port_type == 'input':
                    call_ = descriptor.module.provide_input_port_documentation
                else:
                    raise VistrailsInternalError("Invalid port type")
                doc = call_(port_name)
                
        self.setWindowTitle('Documentation for %s port %s in "%s"' %
                            (port_type, port_name, module.name))


        layout = QtGui.QVBoxLayout()
        layout.addStrut(600)
        layout.addWidget(QtGui.QLabel("Port name: %s" % port_name))
        layout.addWidget(QtGui.QLabel("Module name: %s" % module.name))
        layout.addWidget(QtGui.QLabel("Module package: %s" % \
                                                 module.package))
        self.textEdit = QtGui.QTextEdit(self)
        layout.addWidget(self.textEdit, 1)
        if doc:
            self.textEdit.insertPlainText(doc)
        else:
            self.textEdit.insertPlainText("(Documentation not available)")
        self.textEdit.setReadOnly(True)
        self.textEdit.setTextCursor(QtGui.QTextCursor(self.textEdit.document()))
        self.buttonBox = QtGui.QDialogButtonBox(QtGui.QDialogButtonBox.Ok)
        layout.addWidget(self.buttonBox)
        self.setLayout(layout)
        self.connect(self.buttonBox, QtCore.SIGNAL('accepted()'), self.accept)
